from __future__ import annotations
from typing import Sequence, Tuple
import numpy as np
import pyqtgraph as pg
from psygnal import Signal
from .._shared_utils import convert_color_code, to_rgba


class GraphicComponent:
    native: pg.GraphicsObject

    @property
    def visible(self) -> bool:
        """Linear region visibility."""
        return self.native.isVisible()

    @visible.setter
    def visible(self, value: bool):
        self.native.setVisible(value)

    def update(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self.__class__, k):
                raise AttributeError(f"Cannot set attribute {k} to TextOverlay.")
        for k, v in kwargs.items():
            setattr(self, k, v)


class Region(GraphicComponent):
    """A linear region with magicgui-like API"""

    changed = Signal(Tuple[float, float])
    native: pg.LinearRegionItem

    def __init__(self):
        self.native = pg.LinearRegionItem()

        @self.native.sigRegionChanged.connect
        def _(e=None):
            self.changed.emit(self.native.getRegion())

    @property
    def value(self) -> tuple[float, float]:
        """Get the limits of linear region."""
        return self.native.getRegion()

    @value.setter
    def value(self, value: tuple[float, float]):
        self.native.setRegion(value)

    @property
    def enabled(self) -> bool:
        return self.native.movable

    @enabled.setter
    def enabled(self, value: bool):
        self.native.setMovable(value)

    @property
    def color(self):
        return to_rgba(self.native.brush)

    @color.setter
    def color(self, value):
        value = convert_color_code(value)
        self.native.setBrush(value)


class Roi(GraphicComponent):
    native: pg.ROI

    def __init__(self, pos=(0, 0)):
        self.native = pg.ROI(pos)
        self.native.setZValue(10000)
        self.native.addScaleHandle([1, 1], [0, 0])

    @property
    def border(self):
        return to_rgba(self.native.pen)

    @border.setter
    def border(self, value):
        value = convert_color_code(value)
        self.native.setPen(pg.mkPen(value))
        self.native._updateView()


class TextItem(GraphicComponent):
    """A text item with napari-like API."""

    native: pg.TextItem

    def __init__(
        self, text: str, color: Sequence[float] | str, anchor=(0, 0), angle: float = 0
    ) -> None:
        self.native = pg.TextItem(
            text, color=convert_color_code(color), anchor=anchor, angle=angle
        )

    @property
    def color(self):
        rgba = self.native.color.getRgb()
        return np.array(rgba) / 255

    @color.setter
    def color(self, value):
        value = convert_color_code(value)
        self.native.setText(self.text, value)

    @property
    def background_color(self):
        return to_rgba(self.native.fill)

    @background_color.setter
    def background_color(self, value):
        value = convert_color_code(value)
        self.native.fill = pg.mkBrush(value)
        self.native._updateView()

    @property
    def border(self):
        return to_rgba(self.native.border)

    @border.setter
    def border(self, value):
        value = convert_color_code(value)
        self.native.border = pg.mkPen(value)
        self.native._updateView()

    @property
    def text(self):
        return self.native.toPlainText()

    @text.setter
    def text(self, value: str):
        self.native.setText(value)

    @property
    def pos(self) -> np.ndarray:
        pos = self.native.pos()
        return np.array([pos.x(), pos.y()])

    @pos.setter
    def pos(self, value):
        self.native.setPos(*value)

    @property
    def anchor(self) -> np.ndarray:
        anchor = self.native.anchor
        return np.array([anchor.x(), anchor.y()])

    @anchor.setter
    def anchor(self, value):
        self.native.setAnchor(value)


class ScaleBar(GraphicComponent):
    """A scale bar with napari-like API"""

    native: pg.ScaleBar

    def __init__(self):
        self._unit = ""
        self.native = pg.ScaleBar(10, suffix=self._unit)
        self.native

    @property
    def color(self):
        return to_rgba(self.native.brush)

    @color.setter
    def color(self, value):
        value = convert_color_code(value)
        self.native.brush = pg.mkBrush(value)
        self.native.bar.setBrush(self.native.brush)

    @property
    def unit(self) -> str:
        return self._unit

    @unit.setter
    def unit(self, value: str):
        value = str(value)
        self.native.text.setText(pg.siFormat(self.native.size, suffix=value))
        self._unit = value


class Legend(GraphicComponent):
    native: pg.LegendItem

    def __init__(self, offset=(0, 0)):
        self.native = pg.LegendItem(offset=offset)

    @property
    def color(self):
        """Text color."""
        color = self.native.labelTextColor()
        if color is None:
            color = pg.mkPen(None).color()
        rgba = color.getRgb()
        return np.array(rgba) / 255

    @color.setter
    def color(self, value):
        value = convert_color_code(value)
        self.native.setLabelTextColor(value)

    @property
    def size(self) -> int:
        """Text size."""
        return self.native.labelTextSize()

    @size.setter
    def size(self, value: int):
        self.native.setLabelTextSize(value)

    @property
    def border(self):
        """Border color."""
        return to_rgba(self.native.pen())

    @border.setter
    def border(self, value):
        value = convert_color_code(value)
        self.native.setPen(pg.mkPen(value))

    @property
    def background_color(self):
        """Background color."""
        return to_rgba(self.native.brush())

    @background_color.setter
    def background_color(self, value):
        value = convert_color_code(value)
        self.native.setBrush(pg.mkBrush(value))



from __future__ import annotations
import weakref

from magicgui.widgets import FunctionGui, Image
from magicgui.widgets._bases.widget import Widget
from magicgui.widgets._function_gui import _docstring_to_html

import qtpy
from qtpy.QtWidgets import QWidget, QTreeWidget, QTreeWidgetItem, QSplitter
from qtpy.QtCore import Qt
from typing import Any, Callable, Iterator, TYPE_CHECKING

from .widgets.containers import SplitterContainer
from .widgets.misc import ConsoleTextEdit

from ._gui.mgui_ext import Action, PushButtonPlus, WidgetAction
from ._gui._base import MagicTemplate
from .widgets import DraggableContainer, FreeWidget, Separator
from ._gui.class_gui import (
    CollapsibleClassGui,
    DraggableClassGui,
    ScrollableClassGui,
    ButtonClassGui,
)
from .utils import iter_members, Tooltips, get_signature

if TYPE_CHECKING:
    import numpy as np

# TODO: find, key-binding


class _HelpWidget(QSplitter):
    """
    A Qt widget that will show information of a magic-class widget, built from its
    class structure, function docstrings and type annotations.
    """

    _initial_image_size = 250

    def __init__(self, ui=None, parent=None) -> None:
        super().__init__(orientation=Qt.Horizontal, parent=parent)
        self.setWindowFlag(Qt.Window)
        self._tree = QTreeWidget(self)
        self._tree.itemClicked.connect(self._on_treeitem_clicked)
        self._text = ConsoleTextEdit()
        self._text.read_only = True
        self._mgui_image = Image()
        c = DraggableContainer(widgets=[self._mgui_image])

        def wheelEvent(event):
            ang = event.angleDelta().y()
            v0 = self._mgui_image.min_height
            if ang > 0:
                v = v0 * 1.1
            else:
                v = v0 / 1.1
            self._resize_image(int(v))

        self._mgui_image.native.wheelEvent = wheelEvent

        c.min_height = 120
        self._resize_image(self._initial_image_size)

        widget_right = SplitterContainer(widgets=[c, self._text])

        self.insertWidget(0, self._tree)
        self.insertWidget(1, widget_right.native)

        if ui is not None:
            self.set_tree(ui)
            self._update_ui_view(ui)

        width = self.width()
        left_width = width // 3
        self.setSizes([left_width, width - left_width])

    def set_tree(self, ui: MagicTemplate, root: UiBoundTreeItem = None):
        name = ui.name
        if root is None:
            root = UiBoundTreeItem(self._tree, ui)
            root.setText(0, name)
            root.setExpanded(True)
            self._tree.invisibleRootItem().addChild(root)

        for i, widget in enumerate(_iter_unwrapped_children(ui)):
            if isinstance(widget, MagicTemplate):
                child = UiBoundTreeItem(root, ui=widget)
                self.set_tree(widget, root=child)
                if child.childCount() == 1 and isinstance(
                    child.child(0).ui, MagicTemplate
                ):
                    # If only one magic class is nested, we don't create redundant items
                    grandchild = child.takeChild(0)
                    grandchild.setText(
                        0, f"({i+1}) {widget.name} > {grandchild.ui.name}"
                    )
                    root.removeChild(child)
                    child = grandchild
                else:
                    child.setText(0, f"({i+1}) {widget.name}")

            else:
                if isinstance(getattr(widget, "_inner_widget", widget), Separator):
                    # separator does not need a help
                    continue
                child = UiBoundTreeItem(root, ui=None)
                child.setText(0, f"({i+1}) {widget.name}")

            root.addChild(child)

    def _resize_image(self, v: float):
        self._mgui_image.min_height = v
        self._mgui_image.max_height = v
        self._mgui_image.min_width = v
        self._mgui_image.max_width = v

    def _update_ui_view(self, ui: MagicTemplate):
        img, docs = get_help_info(ui)

        # set image
        self._mgui_image.value = img

        # set text
        htmls = [f"<h1>{ui.name}</h1><p>{Tooltips(ui).desc}</p>"]

        if docs:
            htmls.append("<h2>Contents</h2>")

        for i, (name, doc) in enumerate(docs.items()):
            htmls.append(f"<h3>({i+1}) {name}</h3><p>{doc}</p>")
        self._text.value = "".join(htmls)

    def _on_treeitem_clicked(self, item: UiBoundTreeItem, i: int = 0):
        if item.ui is not None:
            self._update_ui_view(item.ui)
            item.setExpanded(True)
        else:
            self._on_treeitem_clicked(item.parent())
        self._resize_image(self._initial_image_size)


class UiBoundTreeItem(QTreeWidgetItem):
    def __init__(self, parent, ui=None):
        super().__init__(parent)
        if ui is not None:
            self._ui = weakref.ref(ui)
        else:
            self._ui = None

    @property
    def ui(self):
        if self._ui is None:
            return None
        return self._ui()

    def child(self, index: int) -> UiBoundTreeItem:
        # Just for typing
        return super().child(index)

    def takeChild(self, index: int) -> UiBoundTreeItem:
        # Just for typing
        return super().takeChild(index)


class HelpWidget(FreeWidget):
    def __init__(self, ui=None, parent=None):
        super().__init__()
        self._help_widget = _HelpWidget(ui, parent)
        self.set_widget(self._help_widget)
        self.native.setWindowTitle("Help")


def _issubclass(child: Any, parent: Any):
    try:
        return issubclass(child, parent)
    except TypeError:
        return False


def get_help_info(ui: MagicTemplate) -> tuple[np.ndarray, dict[str, str]]:
    import numpy as np
    import matplotlib as mpl
    import matplotlib.pyplot as plt

    backend = mpl.get_backend()
    try:
        _has_inner_widget = (
            ScrollableClassGui,
            DraggableClassGui,
            CollapsibleClassGui,
            ButtonClassGui,
        )
        if isinstance(ui, _has_inner_widget):
            inner_widget = ui._widget._inner_widget
            visible = inner_widget.isVisible()
            inner_widget.setVisible(True)
            img = _render(inner_widget)
            inner_widget.setVisible(visible)
        # elif isinstance(ui, TabbedClassGui):
        # TODO: Cannot assign correct position of tabs now.
        else:
            img = ui.render()
        scale = _screen_scale()
        docs: dict[str, str] = {}
        mpl.use("Agg")
        with plt.style.context("default"):
            fig, ax = plt.subplots(1, 1)
            ax.axis("off")
            fig.patch.set_alpha(0.05)
            ax.imshow(img)
            for i, widget in enumerate(_iter_unwrapped_children(ui)):
                x, y = _get_relative_pos(widget)
                if x >= 0 and y >= 0:
                    ax.text(
                        x * scale,
                        y * scale,
                        f"({i+1})",
                        ha="center",
                        va="center",
                        color="white",
                        backgroundcolor="black",
                        fontfamily="Arial",
                    )
                docs[widget.name] = _get_doc(widget)
            fig.tight_layout()
            fig.canvas.draw()

            data = np.asarray(fig.canvas.renderer.buffer_rgba(), dtype=np.uint8)

    finally:
        mpl.use(backend)

    return data, docs


def _screen_scale() -> float:
    from qtpy.QtGui import QGuiApplication

    screen = QGuiApplication.screens()[0]
    return screen.devicePixelRatio()


def _get_doc(widget) -> str:
    if isinstance(widget, MagicTemplate):
        doc = widget.__doc__ or ""
    elif isinstance(widget, (Action, PushButtonPlus)):
        doc = _docstring_to_html(widget._doc)
    elif isinstance(widget, FunctionGui):
        doc = _docstring_to_html(widget._function.__doc__)
    elif isinstance(widget, (Widget, WidgetAction)):
        doc = widget.tooltip or ""
    else:
        raise TypeError(type(widget))
    doc = doc.rstrip("<h3>Parameters</h3><ul></ul>")  # If parameter info was not given
    if doc == "":
        doc = "(No document found)"
    return doc


def _render(qwidget: QWidget) -> np.ndarray:
    """Render Qt widgets. Used in certain type of containers."""
    import numpy as np

    img = qwidget.grab().toImage()
    bits = img.constBits()
    h, w, c = img.height(), img.width(), 4
    if qtpy.API_NAME == "PySide2":
        arr = np.array(bits).reshape(h, w, c)
    else:
        bits.setsize(h * w * c)
        arr = np.frombuffer(bits, np.uint8).reshape(h, w, c)

    return arr[:, :, [2, 1, 0, 3]]


def _iter_unwrapped_children(ui: MagicTemplate) -> Iterator[Widget]:
    all_widgets: set[Widget] = set()

    for item in ui._list:
        widget = getattr(item, "_inner_widget", item)
        all_widgets.add(widget)
    for widget in ui.__magicclass_children__:
        all_widgets.add(widget)

    for w in all_widgets:
        if not getattr(w, "_unwrapped", False):
            yield w


def _get_relative_pos(widget: Widget) -> tuple[int, int]:
    """Get relative position of a widget seen from its parent."""
    if hasattr(widget, "_labeled_widget"):
        w = widget._labeled_widget()
        if w is None:
            w = widget
        try:
            qpos = w.native.mapToParent(w.native.rect().topLeft())
            out = qpos.x(), qpos.y()
        except Exception:
            out = (-1, -1)
    else:
        out = (-1, -1)
    return out


def get_keymap(ui: MagicTemplate | type[MagicTemplate]):
    from .signature import get_additional_option
    from ._gui.keybinding import as_shortcut

    keymap: dict[str, Callable] = {}

    if isinstance(ui, MagicTemplate):
        cls = ui.__class__
    elif _issubclass(ui, MagicTemplate):
        cls = ui
    else:
        raise TypeError("'get_keymap' can only be called with MagicTemplate input.")

    for name, attr in iter_members(cls, exclude_prefix=" "):
        if isinstance(attr, type) and issubclass(attr, MagicTemplate):
            child_keymap = get_keymap(attr)
            keymap.update(child_keymap)
        else:
            kb = get_additional_option(attr, "keybinding", None)
            if kb:
                keystr = as_shortcut(kb).toString()
                keymap[keystr] = (name, Tooltips(attr).desc)

    return keymap



from __future__ import annotations
import weakref
import numpy as np
import vedo
from qtpy import QtWidgets as QtW
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
from psygnal.containers import EventedList
from .const import AxesMode
from .volume import Volume
from .components import Mesh, VtkComponent, get_object_type

from ...widgets import FreeWidget
from ...types import Color


class QtVtkCanvas(QtW.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        _layout = QtW.QVBoxLayout()
        self.setLayout(_layout)
        self._vtk_widget = QVTKRenderWindowInteractor(parent=self)
        self.plt = vedo.Plotter(qtWidget=self._vtk_widget, bg="black", axes=0)
        self.plt.show()

        _layout.addWidget(self._vtk_widget)


class LayerList(EventedList):
    def __init__(self, data=(), parent: VtkCanvas = None):
        super().__init__(data=data)
        self._parent_ref = weakref.ref(parent)
        self.events.inserted.connect(self._on_inserted)
        self.events.removed.connect(self._on_removed)

    @property
    def parent(self) -> VtkCanvas:
        return self._parent_ref()

    def _on_inserted(self, i: int, obj: VtkComponent):
        self.parent._qwidget.plt.add(obj._obj)
        self.parent._qwidget.plt.window.Render()
        if len(self) == 1:
            self.parent._qwidget.plt.show(zoom=True)

    def _on_removed(self, i: int, obj: VtkComponent):
        self.parent._qwidget.plt.remove(obj._obj)
        self.parent._qwidget.plt.window.Render()
        if len(self) == 1:
            self.parent._qwidget.plt.show(zoom=True)


class VtkCanvas(FreeWidget):
    def __init__(self):
        """
        A Visualization toolkit (VTK) canvas for magicclass.

        This widget is useful for visualizing surface and mesh.
        """
        super().__init__()
        self._qwidget = QtVtkCanvas()
        self.set_widget(self._qwidget)
        self._layers = LayerList(parent=self)

    @property
    def layers(self):
        return self._layers

    def screenshot(self) -> np.ndarray:
        """Get screenshot as a numpy array."""
        pic: vedo.Picture = self._qwidget.plt.topicture()
        img = pic.tonumpy()
        return img

    def add_volume(
        self,
        volume: np.ndarray,
        color: Color = (0.7, 0.7, 0.7),
        mode: str = "iso",
    ):
        """
        Add a 3D volume to the canvas.

        Parameters
        ----------
        volume : np.ndarray
            Volume data. Must be 3D array.
        color : Color, optional
            Initial color of the volume.
        mode : str, default is "iso"
            Initial visualization mode of the volume.

        Returns
        -------
        Volume
            A volume layer.
        """
        vol = Volume(volume, _parent=self._qwidget.plt)
        self.layers.append(vol)
        self._qwidget.plt.add(vol._current_obj)
        vol.color = color
        vol.mode = mode
        if len(self.layers) == 1:
            self._qwidget.plt.show(zoom=True)
        return vol

    def add_object(self, *args, object_type: str = None, **kwargs):
        obj = get_object_type(object_type.capitalize())(
            *args, **kwargs, _parent=self._qwidget.plt
        )
        self.layers.append(obj)
        if len(self.layers) == 1:
            self._qwidget.plt.show()
        return obj

    def add_surface(self, data: tuple[np.ndarray, np.ndarray] | tuple[np.ndarray]):
        mesh = Mesh(data, _parent=self._qwidget.plt)
        self.layers.append(mesh)
        if len(self.layers) == 1:
            self._qwidget.plt.show(zoom=True)
        return mesh

    @property
    def axes(self) -> str:
        """The axes object."""
        return AxesMode(self._qwidget.plt.axes).name

    @axes.setter
    def axes(self, v):
        if self._qwidget.plt.axes_instances:
            current_axes = self._qwidget.plt.axes_instances[0]
        else:
            current_axes = None
        a = getattr(AxesMode, v).value
        try:
            self._qwidget.plt.remove(current_axes)
        except TypeError:
            current_axes.EnabledOff()
        self._qwidget.plt.axes_instances = [None]
        self._qwidget.plt.show(axes=a)



from magicclass import magicclass, magicmenu, set_options, defaults, redo
from enum import Enum
from pathlib import Path
from datetime import datetime, date, time
from unittest.mock import MagicMock


class X(Enum):
    a = 1
    b = 2

class Y(Enum):
    t = "\t"
    n = "\n"

def test_macro_rerun():
    mock = MagicMock()

    @magicclass(error_mode="stderr")
    class A:
        def f0(self, x: X):
            try:
                x = X(x)  # this is the formal way to use Enum
            except Exception:
                mock()

        def f1(self, y: Y):
            try:
                y = Y(y)  # this is the formal way to use Enum
            except Exception:
                mock()

        @set_options(x={"choices": [2, 3]},
                     y={"choices": ["a", "b"]})
        def f2(self, x, y):
            if not isinstance(x, int):
                mock()

            if not isinstance(y, str):
                mock()

        def f3(self, dt: datetime, d: date, t: time):
            try:
                dt.strftime("")
                d.strftime("")
                t.strftime("")
            except Exception:
                mock()

        def f4(self, path: Path, a=1):
            str(path)  # this is the formal way to use Path

    ui = A()
    ui["f0"].changed()
    ui["f0"].mgui[-1].changed()
    ui["f1"].changed()
    ui["f1"].mgui[-1].changed()
    ui["f2"].changed()
    ui["f2"].mgui[-1].changed()
    ui["f3"].changed()
    ui["f3"].mgui[-1].changed()
    ui["f4"].changed()
    ui["f4"].mgui[-1].changed()

    mock.assert_not_called()

    ui.macro.widget.execute_lines(0)
    mock.assert_not_called()
    ui.macro.widget.execute_lines(1)
    mock.assert_not_called()
    ui.macro.widget.execute_lines(2)
    mock.assert_not_called()
    ui.macro.widget.execute_lines(3)
    mock.assert_not_called()
    ui.macro.widget.execute_lines(4)
    mock.assert_not_called()


def test_blocked():
    @magicclass
    class A:
        def f(self, a: int = 1):
            self._a = a
        def g(self):
            return self.f(0)

    ui = A()
    ui["f"].changed()
    ui["f"].mgui[-1].changed()
    assert ui._a == 1
    ui["g"].changed()
    assert ui._a == 0
    ui.f(2)
    assert ui._a == 2
    ui.g()
    assert ui._a == 0
    assert str(ui.macro[1]) == "ui.f(a=1)"
    assert str(ui.macro[2]) == "ui.g()"
    assert str(ui.macro[3]) == "ui.f(a=2)"
    assert str(ui.macro[4]) == "ui.g()"

def test_max_history():
    @magicclass
    class A:
        def f(self): pass
    max_hist = defaults["macro-max-history"]
    ui = A()
    for i in range(max_hist + 10):
        ui.f()
    assert len(ui.macro) == max_hist
    ui.f()
    assert len(ui.macro) == max_hist

def test_init():
    """test macro is blocked during __init__ and __post_init__"""
    @magicclass
    class A:
        @magicmenu
        class Menu:
            def __init__(self):
                self.f()
            def __post_init__(self):
                self.f()
            def f(self): pass

        def __init__(self):
            self.f()
        def __post_init__(self):
            self.f()
        def f(self): pass

    ui = A()
    assert len(ui.macro) == 1

def test_redo():
    @magicclass
    class A:
        def f(self):
            pass
        def g(self, a: int):
            pass
    ui = A()
    ui.f()
    redo(ui)
    ui.g(2)
    redo(ui)
    assert str(ui.macro[1]) == "ui.f()"
    assert str(ui.macro[2]) == "ui.f()"
    assert str(ui.macro[3]) == "ui.g(a=2)"
    assert str(ui.macro[4]) == "ui.g(a=2)"



"""
Advanced widgets for magic class GUI.
These widgets are all compatible with the ``append`` method of Container widgets.
"""

from magicgui.widgets import *  # to avoid importing both magicgui.widgets and magicclass.widgets

from .containers import (
    ButtonContainer,
    GroupBoxContainer,
    FrameContainer,
    ListContainer,
    SubWindowsContainer,
    ScrollableContainer,
    DraggableContainer,
    CollapsibleContainer,
    HCollapsibleContainer,
    SplitterContainer,
    StackedContainer,
    TabbedContainer,
    ToolBoxContainer,
)
from .pywidgets import ListWidget, DictWidget
from .color import ColorEdit, ColorSlider
from .misc import (
    OptionalWidget,
    ConsoleTextEdit,
    CheckButton,
    RangeSlider,
    FloatRangeSlider,
    SpreadSheet,
)
from .plot import Figure, SeabornFigure
from .separator import Separator
from .sequence import ListEdit, TupleEdit
from .utils import FreeWidget
from .logger import Logger

__all__ = [
    "ButtonContainer",
    "ColorEdit",
    "ColorSlider",
    "ConsoleTextEdit",
    "CheckButton",
    "CollapsibleContainer",
    "DictWidget",
    "DraggableContainer",
    "FrameContainer",
    "Figure",
    "FloatRangeSlider",
    "FreeWidget",
    "GroupBoxContainer",
    "HCollapsibleContainer",
    "ListEdit",
    "ListContainer",
    "ListWidget",
    "Logger",
    "OptionalWidget",
    "RangeSlider",
    "SeabornFigure",
    "Separator",
    "ScrollableContainer",
    "SubWindowsContainer",
    "SplitterContainer",
    "SpreadSheet",
    "StackedContainer",
    "TabbedContainer",
    "ToolBoxContainer",
    "TupleEdit",
]



from __future__ import annotations
from typing import Any, Iterable, TypeVar, overload, Iterator, Tuple, Sequence
from typing_extensions import get_args, get_origin, _AnnotatedAlias
import inspect
from magicgui.types import WidgetOptions
from magicgui.signature import split_annotated_type
from magicgui.widgets import create_widget, Container, PushButton, EmptyWidget
from magicgui.widgets._bases.value_widget import UNSET, ValueWidget, _Unset
from magicgui.widgets._concrete import merge_super_sigs

_V = TypeVar("_V")


@merge_super_sigs
class ListEdit(Container):
    """A widget to represent a list of values.
    A ListEdit container can create a list with multiple objects of same type. It
    will contain many child widgets and their value is represented as a Python list
    object. If a list is given as the initial value, types of child widgets are
    determined from the contents. Number of contents can be adjusted with +/-
    buttons.
    Parameters
    ----------
    options: WidgetOptions, optional
        Widget options of child widgets.
    """

    def __init__(
        self,
        value: Iterable[_V] | _Unset = UNSET,
        layout: str = "horizontal",
        labels: bool = False,
        nullable: bool = False,
        options: WidgetOptions = None,
        **kwargs,
    ):
        self._args_type: type | None = None
        self._nullable = nullable
        self._child_options = options or {}

        super().__init__(layout=layout, labels=labels, **kwargs)
        self.margins = (0, 0, 0, 0)

        if not isinstance(value, _Unset):
            types = {type(a) for a in value}
            if len(types) == 1:
                if self._args_type is None:
                    self._args_type = types.pop()
            else:
                raise TypeError("values have inconsistent types.")
            _value: Iterable[_V] = value
        else:
            _value = []

        button_plus = PushButton(text="+", name="plus")
        button_plus.changed.connect(lambda: self._append_value())

        button_minus = PushButton(text="-", name="minus")
        button_minus.changed.connect(self._pop_value)

        if layout == "horizontal":
            button_plus.max_width = 40
            button_minus.max_width = 40

        self.append(button_plus)
        self.append(button_minus)

        for a in _value:
            self._append_value(a)

        self.btn_plus = button_plus
        self.btn_minus = button_minus

    @property
    def annotation(self):
        """Return type annotation for the parameter represented by the widget.
        ForwardRefs will be resolve when setting the annotation. For ListEdit,
        annotation will be like 'list[str]'.
        """
        return self._annotation

    @annotation.setter
    def annotation(self, value):
        if value is None:
            self._annotation = None
            self._args_type = None
            return

        from magicgui._type_wrapper import resolve_annotation

        value = resolve_annotation(value)
        arg: type | None = None

        if value and value is not inspect.Parameter.empty:
            from magicgui.type_map import _is_subclass

            orig = get_origin(value)
            if not (_is_subclass(orig, list) or isinstance(orig, list)):
                raise TypeError(
                    f"cannot set annotation {value} to {type(self).__name__}."
                )
            args = get_args(value)
            if len(args) > 0:
                arg = args[0]
            else:
                arg = None

        if isinstance(arg, _AnnotatedAlias):
            arg, metadata = split_annotated_type(arg)
            if self._child_options and metadata:
                raise ValueError(
                    "Cannot set options from 'options' argument and Annotated type."
                )
            self._child_options = metadata

        self._annotation = value
        self._args_type = arg

    def _append_value(self, value=UNSET):
        """Create a new child value widget and append it."""
        i = len(self) - 2

        widget = create_widget(
            annotation=self._args_type,
            name=f"value_{i}",
            label=f"({i})",
            options=self._child_options,
        )

        if isinstance(widget, EmptyWidget):
            raise TypeError("could not determine the type of child widget.")

        self.insert(i, widget)

        # Value must be set after new widget is inserted because it could be
        # valid only after same parent is shared between widgets.
        if value is UNSET and i > 0:
            value = self[i - 1].value  # type: ignore
        if value is not UNSET:
            widget.value = value

    def _pop_value(self):
        """Delete last child value widget."""
        try:
            self.pop(-3)
        except IndexError:
            pass

    @property
    def buttons_visible(self) -> bool:
        """Toggle visibility of +/- buttons."""
        return self.btn_plus.visible

    @buttons_visible.setter
    def buttons_visible(self, v: bool) -> None:
        self.btn_plus.visible = v
        self.btn_minus.visible = v

    @property
    def value(self) -> list:
        """Return current value as a list object."""
        return list(ListDataView(self))

    @value.setter
    def value(self, vals: Iterable[_V]):
        del self[:-2]
        for v in vals:
            self._append_value(v)

    @property
    def data(self) -> ListDataView:
        """Return a data view of current value."""
        return ListDataView(self)

    @data.setter
    def data(self, vals: Iterable[_V]):
        del self[:-2]
        for v in vals:
            self._append_value(v)


class ListDataView:
    """Data view of ListEdit."""

    def __init__(self, obj: ListEdit):
        self._obj = obj
        self._widgets: list[ValueWidget] = list(obj[:-2])  # type: ignore

    def __repr__(self):
        """A list-like representation."""
        return f"{self.__class__.__name__}({list(self)!r})"

    def __len__(self):
        """Length as a list."""
        return len(self._widgets)

    def __eq__(self, other):
        """Compare as a list."""
        return list(self) == other

    @overload
    def __getitem__(self, i: int) -> _V:  # noqa
        ...

    @overload
    def __getitem__(self, key: slice) -> list[_V]:  # noqa
        ...

    def __getitem__(self, key):
        """Slice as a list."""
        if isinstance(key, int):
            return self._widgets[key].value
        elif isinstance(key, slice):
            return [w.value for w in self._widgets[key]]
        else:
            raise TypeError(
                f"list indices must be integers or slices, not {type(key).__name__}"
            )

    @overload
    def __setitem__(self, key: int, value: _V) -> None:  # noqa
        ...

    @overload
    def __setitem__(self, key: slice, value: _V | Iterable[_V]) -> None:  # noqa
        ...

    def __setitem__(self, key, value):
        """Update widget value."""
        if isinstance(key, int):
            self._widgets[key].value = value
        elif isinstance(key, slice):
            if isinstance(value, type(self._widgets[0].value)):
                for w in self._widgets[key]:
                    w.value = value
            else:
                for w, v in zip(self._widgets[key], value):
                    w.value = v
        else:
            raise TypeError(
                f"list indices must be integers or slices, not {type(key).__name__}"
            )

    @overload
    def __delitem__(self, key: int) -> None:  # noqa
        ...

    @overload
    def __delitem__(self, key: slice) -> None:  # noqa
        ...

    def __delitem__(self, key):
        """Delete widget at the key(s)."""
        self._obj.__delitem__(key)

    def __iter__(self) -> Iterator[_V]:
        """Iterate over values of child widgets."""
        for w in self._widgets:
            yield w.value


@merge_super_sigs
class TupleEdit(Container):
    """A widget to represent a tuple of values.
    A TupleEdit container has several child widgets of different type. Their value is
    represented as a Python tuple object. If a tuple is given as the initial value,
    types of child widgets are determined one by one. Unlike ListEdit, number of
    contents is not editable.
    Parameters
    ----------
    options: WidgetOptions, optional
        Widget options of child widgets.
    """

    def __init__(
        self,
        value: Iterable[_V] | _Unset = UNSET,
        layout: str = "horizontal",
        nullable: bool = False,
        options: WidgetOptions = None,
        **kwargs,
    ):
        self._nullable = nullable
        self._args_types: tuple[type, ...] | None = None
        super().__init__(layout=layout, labels=False, **kwargs)
        self._child_options = options or {}
        self.margins = (0, 0, 0, 0)

        if not isinstance(value, _Unset):
            if self._args_types is None:
                self._args_types = tuple(type(a) for a in value)
            _value: Iterable[Any] = value
        elif self._args_types is not None:
            _value = (UNSET,) * len(self._args_types)
        else:
            raise ValueError(
                "Either 'value' or 'annotation' must be specified in "
                f"{type(self).__name__}."
            )

        for i, a in enumerate(_value):
            annot = self._args_types[i]
            if isinstance(annot, _AnnotatedAlias):
                annot, metadata = split_annotated_type(annot)
                options = self._child_options.copy()
                options.update(metadata)
            else:
                options = self._child_options
            widget = create_widget(
                value=a,
                annotation=annot,
                name=f"value_{i}",
                options=options,
            )
            self.insert(i, widget)

    def __iter__(self) -> Iterator[ValueWidget]:
        """Just for typing."""
        return super().__iter__()  # type: ignore

    @property
    def annotation(self):
        """Return type annotation for the parameter represented by the widget.
        ForwardRefs will be resolve when setting the annotation. For TupleEdit,
        annotation will be like 'tuple[str, int]'.
        """
        return self._annotation

    @annotation.setter
    def annotation(self, value):
        if value is None:
            self._annotation = None
            self._args_types = None
            return
        from magicgui._type_wrapper import resolve_annotation

        value = resolve_annotation(value)
        args: tuple[type, ...] | None = None

        if value and value is not inspect.Parameter.empty:
            from magicgui.type_map import _is_subclass

            orig = get_origin(value)
            if not (_is_subclass(orig, tuple) or isinstance(orig, tuple)):
                raise TypeError(
                    f"cannot set annotation {value} to {type(self).__name__}."
                )
            args = get_args(value)
            value = Tuple[args]

        self._annotation = value
        self._args_types = args

    @property
    def value(self) -> tuple:
        """Return current value as a tuple."""
        return tuple(w.value for w in self)

    @value.setter
    def value(self, vals: Sequence):
        if len(vals) != len(self):
            raise ValueError("Length of tuple does not match.")

        for w, v in zip(self, vals):
            w.value = v  # type: ignore



from __future__ import annotations
from typing import Callable

__all__ = [
    "napari_light_theme",
    "napari_dark_theme",
    "napari_system_theme",
]


class StyleSheet:
    def __init__(self, getter: Callable[[], str]):
        self.getter = getter

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<{self.getter.__name__}>"

    def __str__(self) -> str:
        """Return style sheet."""
        return self.getter()

    def __add__(self, other: StyleSheet) -> StyleSheet:
        """Conjugate stylesheets."""

        def new_getter():
            s0 = self.getter()
            s1 = other.getter()
            return s0 + "\n" + s1

        return new_getter


def _napari_get_stylesheet(theme: str):
    # TODO: don't launch a viewer if possible.
    import napari

    viewer = napari.Viewer(show=False)
    viewer.theme = theme
    stylesheet = viewer.window._qt_window.styleSheet()
    viewer.close()
    return stylesheet


@StyleSheet
def napari_light_theme():
    return _napari_get_stylesheet("light")


@StyleSheet
def napari_dark_theme():
    return _napari_get_stylesheet("dark")


@StyleSheet
def napari_system_theme():
    return _napari_get_stylesheet("system")



from magicclass import field, magicclass
from magicclass.ext.pyqtgraph import QtPlotCanvas
import numpy as np

def test_plot_canvas():
    @magicclass
    class A:
        plot = field(QtPlotCanvas)

    ui = A()
    ndata = 100
    data = np.random.random(ndata)

    # test curve
    ui.plot.add_curve(data)
    ui.plot.add_curve(np.arange(ndata)*2, data)
    item = ui.plot.add_curve(data, face_color="green", edge_color="green", name="test", lw=2, ls=":")
    item.visible
    item.visible = False
    item.face_color
    item.face_color = "yellow"
    item.edge_color
    item.edge_color = "yellow"
    item.lw
    item.lw = 1
    item.ls
    for ls in ["-", "--", ":", "-."]:
        item.ls = ls
    item.symbol
    for sym in ["o", "s", "D", "^", "<", "v", ">", "*"]:
        item.symbol = sym
    item.xdata = item.xdata + 1
    item.ydata = item.ydata + 1
    assert item.ndata == ndata

    assert len(ui.plot.layers) == 3
    ui.plot.layers.clear()

    # test scatter
    ui.plot.add_scatter(data)
    ui.plot.add_scatter(np.arange(ndata)*2, data)
    item = ui.plot.add_scatter(data, face_color="green", edge_color="green", name="test", lw=2, ls=":", size=6, symbol="*")
    item.visible
    item.visible = False
    item.face_color
    item.face_color = "yellow"
    item.edge_color
    item.edge_color = "yellow"
    item.lw
    item.lw = 1
    item.ls
    for ls in ["-", "--", ":", "-."]:
        item.ls = "-"
    for sym in ["o", "s", "D", "^", "<", "v", ">", "*"]:
        item.symbol = sym
    item.xdata = item.xdata + 1
    item.ydata = item.ydata + 1
    assert item.ndata == ndata

    assert len(ui.plot.layers) == 3
    ui.plot.layers.clear()

    # test bar
    ui.plot.add_bar(data)
    ui.plot.add_bar(np.arange(ndata)*2, data)
    item = ui.plot.add_bar(data, face_color="green", edge_color="green", name="test", lw=2, ls=":")
    item.visible
    item.visible = False
    item.face_color
    item.face_color = "yellow"
    item.edge_color
    item.edge_color = "yellow"
    item.lw
    item.lw = 1
    item.ls
    for ls in ["-", "--", ":", "-."]:
        item.ls = ls
    item.xdata = item.xdata + 1
    item.ydata = item.ydata + 1
    assert item.ndata == ndata

    assert len(ui.plot.layers) == 3
    ui.plot.layers.clear()

    ui.plot.add_hist(data)
    item = ui.plot.add_hist(data, bins=14, range=[-0.5, 1.5], density=True)
    item.visible
    item.visible = False
    item.face_color
    item.face_color = "yellow"
    item.edge_color
    item.edge_color = "yellow"
    item.lw
    item.lw = 1
    item.ls
    for ls in ["-", "--", ":", "-."]:
        item.ls = ls

    assert len(ui.plot.layers) == 2

    # test region
    ui.plot.region.visible
    ui.plot.region.visible = True
    ui.plot.region.value
    ui.plot.region.value = [-1, 2]
    ui.plot.region.enabled
    ui.plot.region.enabled = True
    ui.plot.region.color
    ui.plot.region.color = "blue"

    # test legend
    ui.plot.legend.visible
    ui.plot.legend.visible = True
    ui.plot.legend.color
    ui.plot.legend.color = "black"
    ui.plot.legend.size
    ui.plot.legend.size = 10
    ui.plot.legend.background_color
    ui.plot.legend.background_color = [1, 1, 1, 0.2]
    ui.plot.legend.border
    ui.plot.legend.border = "red"



import re

shared_docs = {
    "x": """
        x : array-like
            X data.""",
    "y": """
        y : array-like
            Y data.""",
    "face_color": """
        face_color: str or array-like, optional
            Face color of plot. Graphic object will be filled with this color.""",
    "edge_color": """
        edge_color: str or array-like, optional
            Edge color of plot.""",
    "color": """
        color: str or array-like, optional
            Set face color and edge color at the same time.""",
    "name": """
        name: str, optional
            Object name of the plot item.""",
    "lw": """
        lw: float, default is 1.0
            Line width of edge.""",
    "ls": """
        ls: str, default is "-"
            Line style of edge. One of "-", "--", ":" or "-.".""",
    "symbol": """
        symbol: str, optional
            Symbol style. Currently supports circle ("o"), cross ("+", "x"), star ("*"),
            square ("s", "D") triangle ("^", "<", "v", ">") and others that ``pyqtgraph``
            supports.""",
}


def write_docs(func):
    doc = func.__doc__
    if doc is not None:
        summary, params, rest = _split_doc(doc)
        for key, value in shared_docs.items():
            value = value.strip()
            params = re.sub("{" + key + "}", value, params)
        doc = _merge_doc(summary, params, rest)
        func.__doc__ = doc
    return func


def _split_doc(doc: str):
    summary, other = doc.split("Parameters\n")
    params, rest = other.split("Returns\n")
    return summary, params, rest


def _merge_doc(summary, params, rest):
    return summary + "Parameters\n" + params + "Returns\n" + rest



from __future__ import annotations
from typing import Any, TypeVar, Callable
import warnings
from qtpy import QtWidgets as QtW
from qtpy.QtCore import Qt, QEvent, QSize
from magicgui.application import use_app
from magicgui.widgets._bases import Widget
from magicgui.widgets._concrete import ContainerWidget
from magicgui.backends._qtpy.widgets import (
    QBaseWidget,
    Container as ContainerBase,
    MainWindow as MainWindowBase,
)

from .utils import merge_super_sigs

# Container variations that is useful in making GUI designs better.

C = TypeVar("C", bound=ContainerWidget)


def wrap_container(cls: type[C] = None, base: type = None) -> Callable | type[C]:
    """
    Provide a wrapper for a new container widget with a new protocol.
    """

    def wrapper(cls) -> type[Widget]:
        def __init__(self, **kwargs):
            app = use_app()
            assert app.native
            kwargs["widget_type"] = base
            super(cls, self).__init__(**kwargs)

        cls.__init__ = __init__
        cls = merge_super_sigs(cls)
        return cls

    return wrapper(cls) if cls else wrapper


def _btn_text_warning():
    msg = "'btn_text' is deprecated and will be removed soon. Please use 'text'."
    warnings.warn(msg, DeprecationWarning)


class _Splitter(ContainerBase):
    _qwidget: QtW.QWidget

    def __init__(self, layout="vertical", scrollable: bool = False):
        QBaseWidget.__init__(self, QtW.QWidget)
        # SetLayout is not supported for QSplitter.
        # Layout is just a dummy.
        self._splitter = QtW.QSplitter(self._qwidget)
        if layout == "horizontal":
            self._splitter.setOrientation(Qt.Horizontal)
            self._layout = QtW.QHBoxLayout()
        else:
            self._splitter.setOrientation(Qt.Vertical)
            self._layout = QtW.QVBoxLayout()

        self._scroll_area = None
        self._qwidget.setLayout(QtW.QVBoxLayout())
        self._qwidget.layout().addWidget(self._splitter)
        self._qwidget.layout().setContentsMargins(0, 0, 0, 0)

    def _mgui_insert_widget(self, position: int, widget: Widget):
        self._splitter.insertWidget(position, widget.native)

    def _mgui_remove_widget(self, widget: Widget):
        widget.native.setParent(None)

    def _mgui_get_margins(self) -> tuple[int, int, int, int]:
        return (0, 0, 0, 0)

    def _mgui_set_margins(self, margins: tuple[int, int, int, int]) -> None:
        pass


class _ToolBox(ContainerBase):
    _qwidget: QtW.QToolBox

    def __init__(self, layout="vertical", scrollable: bool = False):
        QBaseWidget.__init__(self, QtW.QToolBox)

        if layout == "horizontal":
            msg = "Horizontal ToolBox is not implemented yet."
            warnings.warn(msg, UserWarning)

        self._layout = self._qwidget.layout()

    def _mgui_insert_widget(self, position: int, widget: Widget):
        self._qwidget.insertItem(position, widget.native, widget.name)

    def _mgui_remove_widget(self, widget: Widget):
        for i in range(self._qwidget.count()):
            if self._qwidget.widget(i) is widget.native:
                self._qwidget.removeItem(i)
                widget.native.setParent(None)
                break
        else:
            raise ValueError(f"Widget {widget.name} not found.")


class _Tab(ContainerBase):
    def __init__(self, layout="vertical", scrollable: bool = False):
        QBaseWidget.__init__(self, QtW.QWidget)

        if layout == "horizontal":
            self._layout: QtW.QLayout = QtW.QHBoxLayout()
        else:
            self._layout = QtW.QVBoxLayout()
        self._scroll_area = None
        self._tab_widget = QtW.QTabWidget(self._qwidget)
        self._tab_widget.setLayout(self._layout)
        self._qwidget.setLayout(QtW.QVBoxLayout())
        self._qwidget.layout().addWidget(self._tab_widget)
        self._qwidget.layout().setContentsMargins(0, 0, 0, 0)

    def _mgui_insert_widget(self, position: int, widget: Widget):
        self._tab_widget.insertTab(position, widget.native, widget.name)

    def _mgui_remove_widget(self, widget: Widget):
        for i in range(self._tab_widget.count()):
            if self._tab_widget.widget(i) is widget.native:
                self._tab_widget.removeTab(i)
                widget.native.setParent(None)
                break
        else:
            raise ValueError(f"Widget {widget.name} not found.")


class _Stack(ContainerBase):
    def __init__(self, layout="vertical", scrollable: bool = False):
        QBaseWidget.__init__(self, QtW.QWidget)

        if layout == "horizontal":
            self._layout: QtW.QLayout = QtW.QHBoxLayout()
        else:
            self._layout = QtW.QVBoxLayout()

        self._stacked_widget = QtW.QStackedWidget(self._qwidget)
        self._stacked_widget.setContentsMargins(0, 0, 0, 0)
        self._inner_widget = QtW.QWidget(self._qwidget)
        self._qwidget.setLayout(self._layout)
        self._layout.addWidget(self._stacked_widget)
        self._layout.addWidget(self._inner_widget)

    def _mgui_insert_widget(self, position: int, widget: Widget):
        self._stacked_widget.insertWidget(position, widget.native)

    def _mgui_remove_widget(self, widget: Widget):
        self._stacked_widget.removeWidget(widget.native)
        widget.native.setParent(None)


class _ScrollableContainer(ContainerBase):
    def __init__(self, layout="vertical", scrollable: bool = False):
        QBaseWidget.__init__(self, QtW.QWidget)
        self._scroll_area = QtW.QScrollArea(self._qwidget)
        if layout == "horizontal":
            self._layout: QtW.QLayout = QtW.QHBoxLayout()
            self._layout.setAlignment(Qt.AlignCenter)
        else:
            self._layout = QtW.QVBoxLayout()
            self._layout.setAlignment(Qt.AlignTop)

        self._scroll_area.setWidgetResizable(True)
        self._scroll_area.setContentsMargins(0, 0, 0, 0)
        self._inner_widget = QtW.QWidget(self._scroll_area)
        self._inner_widget.setLayout(self._layout)
        self._scroll_area.setWidget(self._inner_widget)

        self._qwidget.setLayout(QtW.QVBoxLayout())
        self._qwidget.layout().addWidget(self._scroll_area)
        self._qwidget.layout().setContentsMargins(0, 0, 0, 0)


class _WheelDisabledScrollArea(QtW.QScrollArea):
    def eventFilter(self, source, event: QEvent):
        if event.type() == QEvent.Wheel:
            return True
        return super().eventFilter(source, event)


class _DraggableContainer(ContainerBase):
    def __init__(self, layout="vertical", scrollable: bool = False):
        QBaseWidget.__init__(self, QtW.QWidget)
        self._scroll_area = _WheelDisabledScrollArea(self._qwidget)
        if layout == "horizontal":
            self._layout: QtW.QLayout = QtW.QHBoxLayout()
            self._layout.setAlignment(Qt.AlignCenter)
        else:
            self._layout = QtW.QVBoxLayout()
            self._layout.setAlignment(Qt.AlignTop)

        self._scroll_area.setWidgetResizable(True)
        self._scroll_area.setContentsMargins(0, 0, 0, 0)
        self._inner_widget = QtW.QWidget(self._scroll_area)
        self._inner_widget.setLayout(self._layout)
        self._scroll_area.setWidget(self._inner_widget)

        self._qwidget.setLayout(QtW.QVBoxLayout())
        self._qwidget.layout().addWidget(self._scroll_area)
        self._qwidget.layout().setContentsMargins(0, 0, 0, 0)
        QtW.QScroller.grabGesture(
            self._scroll_area, QtW.QScroller.LeftMouseButtonGesture
        )

        self._scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self._scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)


class _ButtonContainer(ContainerBase):
    def __init__(self, layout="vertical", text="", scrollable: bool = False):
        QBaseWidget.__init__(self, QtW.QWidget)
        if layout == "horizontal":
            self._layout: QtW.QLayout = QtW.QHBoxLayout()
        else:
            self._layout = QtW.QVBoxLayout()

        self._qwidget = QtW.QPushButton()
        self._inner_widget = QtW.QWidget()
        self._inner_widget.setParent(self._qwidget, self._inner_widget.windowFlags())
        self._inner_widget.setLayout(self._layout)

        self._qwidget.setText(text)
        self._qwidget.clicked.connect(lambda x: self._inner_widget.show())


_VERTICAL_SETTING = {
    "expanded-arrow": Qt.ArrowType.DownArrow,
    "collapsed-arrow": Qt.ArrowType.RightArrow,
    "align": Qt.AlignTop,
    "text-align": "left",
    "layout": QtW.QVBoxLayout,
}
_HORIZONTAL_SETTING = {
    "expanded-arrow": Qt.ArrowType.RightArrow,
    "collapsed-arrow": Qt.ArrowType.LeftArrow,
    "align": Qt.AlignLeft,
    "text-align": "center",
    "layout": QtW.QHBoxLayout,
}


class _QCollapsible(QtW.QWidget):
    def sizeHint(self) -> QSize:
        if self.isVisible():
            return super().sizeHint()
        else:
            return QSize(0, 0)


class _Collapsibles(ContainerBase):
    _setting: dict[str, Any]

    def __init__(self, layout="vertical", text="", scrollable: bool = False):
        QBaseWidget.__init__(self, QtW.QWidget)
        if layout == "horizontal":
            self._layout: QtW.QLayout = QtW.QHBoxLayout()
        else:
            self._layout = QtW.QVBoxLayout()

        self._get_setting()
        self._qwidget = QtW.QWidget()
        self._qwidget.setLayout(self._setting["layout"]())
        self._inner_widget = _QCollapsible(self._qwidget)
        self._inner_widget.setLayout(self._layout)

        self._qwidget.layout().setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)

        self._expand_btn = QtW.QToolButton(self._qwidget)
        self._expand_btn.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self._expand_btn.setArrowType(self._setting["collapsed-arrow"])
        self._expand_btn.setText(text)
        self._expand_btn.setCheckable(True)
        self._expand_btn.setChecked(False)
        self._expand_btn.setStyleSheet(
            f"""
            QToolButton {{
                border: none;
                text-align: {self._setting['text-align']};
                }}
            """
        )
        self._expand_btn.clicked.connect(self._mgui_change_expand)
        self._mgui_change_expand()

        self._qwidget.layout().addWidget(self._expand_btn, 0, self._setting["align"])
        self._qwidget.layout().addWidget(self._inner_widget, 0, self._setting["align"])
        self._qwidget.layout().setContentsMargins(0, 0, 0, 0)

    @property
    def collapsed(self) -> bool:
        return not self._expand_btn.isChecked()

    def _collapse(self):
        self._inner_widget.setVisible(False)
        self._expand_btn.setArrowType(self._setting["collapsed-arrow"])

    def _expand(self):
        self._inner_widget.setVisible(True)
        self._expand_btn.setArrowType(self._setting["expanded-arrow"])

    def _mgui_change_expand(self):
        if self.collapsed:
            self._collapse()
        else:
            self._expand()

    def _get_setting(self):
        raise NotImplementedError()


class _VCollapsibleContainer(_Collapsibles):
    def _get_setting(self):
        self._setting = _VERTICAL_SETTING


class _HCollapsibleContainer(_Collapsibles):
    def _get_setting(self):
        self._setting = _HORIZONTAL_SETTING


class _QListWidget(QtW.QListWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAcceptDrops(True)
        self.setDragEnabled(True)
        self.setDragDropMode(QtW.QAbstractItemView.DragDropMode.InternalMove)

    def dropEvent(self, event):
        widget = self.itemWidget(self.currentItem())
        super().dropEvent(event)

        # It seems that QListView has a bug in drag-and-drop.
        # When we tried to move the first item to the upper half region of the
        # second item, the inner widget of the first item dissapears.
        # This bug seemed to be solved to set the inner widget again.
        item = self.itemAt(event.pos())
        dest = self.itemWidget(item)
        if dest is None:
            self.setItemWidget(item, widget)


class _ListContainer(ContainerBase):
    def __init__(self, layout="vertical", scrollable: bool = False):
        QBaseWidget.__init__(self, QtW.QWidget)
        self._listwidget = _QListWidget(self._qwidget)
        if layout == "horizontal":
            self._layout: QtW.QLayout = QtW.QHBoxLayout()
        else:
            self._layout = QtW.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._listwidget)
        self._qwidget.setLayout(self._layout)

        if layout == "horizontal":
            self._listwidget.setFlow(QtW.QListView.LeftToRight)
        else:
            self._listwidget.setFlow(QtW.QListView.TopToBottom)

    def _mgui_insert_widget(self, position: int, widget: Widget):
        item = QtW.QListWidgetItem(self._listwidget)
        item.setSizeHint(widget.native.sizeHint())
        self._listwidget.insertItem(position, item)
        self._listwidget.setItemWidget(item, widget.native)

    def _mgui_remove_widget(self, widget: Widget):
        for i in range(self._listwidget.count()):
            item = self._listwidget.item(i)
            w = self._listwidget.itemWidget(item)
            if widget.native is w:
                self._listwidget.removeItemWidget(item)
                self._listwidget.takeItem(i)
                break
        else:
            raise ValueError(f"Widget {widget} not found in the list.")

        widget.native.setParent(None)


class _SubWindowsContainer(ContainerBase):
    # The close button in QMdiArea completely deletes the sub window widget. This accident
    # can be avoided by defining a custom window flag.
    _NoCloseButtonFlag = (
        Qt.CustomizeWindowHint | Qt.WindowTitleHint | Qt.WindowMinMaxButtonsHint
    )

    def __init__(self, layout="vertical", scrollable: bool = False):
        QBaseWidget.__init__(self, QtW.QWidget)
        self._mdiarea = QtW.QMdiArea(self._qwidget)
        if layout == "horizontal":
            self._layout: QtW.QLayout = QtW.QHBoxLayout()
        else:
            self._layout = QtW.QVBoxLayout()

        self._layout.addWidget(self._mdiarea)
        self._qwidget.setLayout(self._layout)

    def _mgui_insert_widget(self, position: int, widget: Widget):
        # position does not have any effect
        self._mdiarea.addSubWindow(widget.native, self._NoCloseButtonFlag)

    def _mgui_remove_widget(self, widget: Widget):
        self._mdiarea.removeSubWindow(widget.native)
        widget.native.setParent(None)


class _GroupBoxContainer(ContainerBase):
    def __init__(self, layout="vertical", scrollable: bool = False):
        QBaseWidget.__init__(self, QtW.QWidget)

        # To precisely control margins, _layout should not be set to the QGroupBox widget.
        self._groupbox = QtW.QGroupBox(self._qwidget)
        if layout == "horizontal":
            self._layout: QtW.QLayout = QtW.QHBoxLayout()
        else:
            self._layout = QtW.QVBoxLayout()

        self._inner_widget = QtW.QWidget(self._groupbox)
        self._inner_widget.setLayout(self._layout)
        self._groupbox.setLayout(QtW.QHBoxLayout())
        self._groupbox.layout().addWidget(self._inner_widget)

        self._qwidget.setLayout(QtW.QVBoxLayout())
        self._qwidget.layout().addWidget(self._groupbox)
        self._qwidget.layout().setContentsMargins(0, 0, 0, 0)

    def _set_title(self, title: str):
        self._groupbox.setTitle(title)


class _FrameContainer(_GroupBoxContainer):
    def __init__(self, layout="vertical", scrollable: bool = False):
        super().__init__(layout=layout)
        self._groupbox.setTitle("")


# Container Widgets


@wrap_container(base=_Splitter)
class SplitterContainer(ContainerWidget):
    """A Container equipped with splitter"""


@wrap_container(base=_ToolBox)
class ToolBoxContainer(ContainerWidget):
    """A Tool box Widget."""

    @property
    def current_index(self):
        return self.native.currentIndex()

    @current_index.setter
    def current_index(self, index: int):
        self.native.setCurrentIndex(index)


@wrap_container(base=_Tab)
class TabbedContainer(ContainerWidget):
    """A tab categorized Container Widget."""

    @property
    def current_index(self):
        return self._widget._tab_widget.currentIndex()

    @current_index.setter
    def current_index(self, index: int):
        self._widget._tab_widget.setCurrentIndex(index)


@wrap_container(base=_Stack)
class StackedContainer(ContainerWidget):
    """A stacked Container Widget"""

    @property
    def current_index(self):
        return self._widget._stacked_widget.currentIndex()

    @current_index.setter
    def current_index(self, index: int):
        self._widget._stacked_widget.setCurrentIndex(index)


@wrap_container(base=_ScrollableContainer)
class ScrollableContainer(ContainerWidget):
    """A scrollable Container Widget."""


@wrap_container(base=_DraggableContainer)
class DraggableContainer(ContainerWidget):
    """A draggable Container Widget."""


@wrap_container(base=_ButtonContainer)
class ButtonContainer(ContainerWidget):
    """A Container Widget hidden in a button."""

    @property
    def btn_text(self):
        _btn_text_warning()
        return self._widget._qwidget.text()

    @btn_text.setter
    def btn_text(self, text: str):
        _btn_text_warning()
        self._widget._qwidget.setText(text)

    @property
    def text(self):
        return self._widget._qwidget.text()

    @text.setter
    def text(self, text: str):
        self._widget._qwidget.setText(text)


@wrap_container(base=_VCollapsibleContainer)
class CollapsibleContainer(ContainerWidget):
    """A collapsible Container Widget."""

    @property
    def btn_text(self):
        _btn_text_warning()
        return self._widget._expand_btn.text()

    @btn_text.setter
    def btn_text(self, text: str):
        _btn_text_warning()
        self._widget._expand_btn.setText(text)

    @property
    def text(self):
        return self._widget._expand_btn.text()

    @text.setter
    def text(self, text: str):
        self._widget._expand_btn.setText(text)

    @property
    def collapsed(self) -> bool:
        return self._widget.collapsed

    @collapsed.setter
    def collapsed(self, value: bool):
        if value:
            self._widget._collapse()
        else:
            self._widget._expand()


@wrap_container(base=_HCollapsibleContainer)
class HCollapsibleContainer(ContainerWidget):
    """A collapsible Container Widget."""

    @property
    def collapsed(self) -> bool:
        return self._widget.collapsed

    @collapsed.setter
    def collapsed(self, value: bool):
        if value:
            self._widget._collapse()
        else:
            self._widget._expand()


@wrap_container(base=_ListContainer)
class ListContainer(ContainerWidget):
    """A Container Widget that support drag and drop."""

    def _post_init(self):
        self._widget._listwidget.model().rowsMoved.connect(self._drag_and_drop_happened)
        return super()._post_init()

    def _drag_and_drop_happened(self, e=None):
        """Sort widget list when item drag/drop happens."""
        l = []
        for i in range(self._widget._listwidget.count()):
            item = self._widget._listwidget.item(i)
            w = self._widget._listwidget.itemWidget(item)

            for widget in self._list:
                # if widget is a _LabeledWidget, inserted widget item will
                # be the parent of the widget in "_list".
                if widget.native is w or widget.native.parent() is w:
                    l.append(widget)
                    break

        self._list = l

    @property
    def current_index(self):
        return self._widget._listwidget.currentRow()

    @current_index.setter
    def current_index(self, index: int):
        self._widget._listwidget.setCurrentRow(index)


@wrap_container(base=_SubWindowsContainer)
class SubWindowsContainer(ContainerWidget):
    """A window-in-window container"""


@wrap_container(base=_GroupBoxContainer)
class GroupBoxContainer(ContainerWidget):
    """A QGroupBox like container"""

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value
        self._widget._set_title(value.replace("_", " "))


@wrap_container(base=_FrameContainer)
class FrameContainer(ContainerWidget):
    """A QGroupBox like container without title."""



"""
This widget submodule contains "visible PyObject" widgets, which has similar API as PyObject to
operate items. They are also equipped with custom delegate, contextmenu and double-click
callbacks.

ListWidget
----------

``ListWidget`` is a ``QListWidget`` wrapper class. This widget can contain any Python objects as
list items.

.. code-block:: python

    from magicclass.widgets import ListWidget

    listwidget = ListWidget()

    # You can add any objects
    listwidget.append("abc")
    listwidget.append(np.arange(5))

You can dispatch double click callbacks depending on the type of contents.

.. code-block:: python

    @listwidget.register_callback(str)
    def _(item, i):
        # This function will be called when the item "abc" is double-clicked.
        print(item)

    @listwidget.register_callback(np.ndarray)
    def _(item, i):
        # This function will be called when the item np.arange(5) is double-clicked.
        print(item.tolist())

In a similar way, you can dispatch display method and context menu.

.. code-block:: python

    @listwidget.register_delegate(np.ndarray)
    def _(item):
        # This function should return how ndarray will be displayed.
        return f"Array with shape {item.shape}"

    @listwidget.register_contextmenu(np.ndarray)
    def Plot(item, i):
        '''Function documentation will be the tooltip.'''
        plt.plot(item)
        plt.show()

DictWidget
----------

DictWidget is a single column QTableWidget. This widget can contain any Python objects
as table items and keys as row names..

.. code-block:: python

    from magicclass.widgets import DictWidget

    dictwidget = DictWidget()

    # You can add any objects
    dictwidget["name-1"] = 10
    dictwidget["data"] = np.arange(5)

The dispatching feature is shared with ListWidget.

"""

from .dict import DictWidget
from .list import ListWidget

__all__ = ["DictWidget", "ListWidget"]



from magicclass import magicclass, magicmenu, MagicTemplate, field, vfield
from unittest.mock import MagicMock
import pytest

def test_simple_callback():
    mock_x = MagicMock()
    mock_y = MagicMock()

    @magicclass
    class A(MagicTemplate):
        x = field(int)
        y = vfield(str)

        @x.connect
        def _callback_x(self):
            mock_x()

        @y.connect
        def _callback_y(self):
            mock_y()

    ui = A()
    ui.x.value
    ui.y
    mock_x.assert_not_called()
    mock_y.assert_not_called()
    ui.x.value += 1
    mock_x.assert_called_once()
    mock_y.assert_not_called()
    ui.y = "y"
    mock_x.assert_called_once()
    mock_y.assert_called_once()

def test_macro_blocked():
    @magicclass
    class A(MagicTemplate):
        x = field(int)
        y = vfield(str)
        result = field(str)

        @x.connect
        def _callback_x(self):
            self.some_function()
            self.result.value = "x changed"

        @y.connect
        def _callback_y(self):
            self.some_function()
            self.result.value = "y changed"

        def some_function(self):
            pass

    ui = A()
    ui.x.value = 1
    assert ui.result.value == "x changed"
    assert str(ui.macro[-1]) == "ui.x.value = 1"
    ui.y = "y"
    assert ui.result.value == "y changed"
    assert str(ui.macro[-1]) == "ui.y = 'y'"


def test_callback_in_parent():
    mock = MagicMock()
    mock2 = MagicMock()

    @magicclass
    class A(MagicTemplate):
        @magicclass
        class B(MagicTemplate):
            b_x = field(int)
            b_y = vfield(int)
            @magicmenu
            class M(MagicTemplate):
                m_x = field(int)
                m_y = vfield(int)

            @M.m_x.connect
            def _callback_m_x(self):
                mock(name="m_x/B")

            @M.m_y.connect
            def _callback_m_y(self):
                mock(name="m_y/B")

        @B.M.m_x.connect
        def _callback_m_x(self):
            mock2(name="m_x/A")

        @B.M.m_y.connect
        def _callback_m_y(self):
            mock2(name="m_y/A")

        @B.b_x.connect
        def _callback_b_x(self):
            mock(name="b_x/A")

        @B.b_y.connect
        def _callback_b_y(self):
            mock(name="b_y/A")

    ui = A()
    mock.assert_not_called()
    ui.B.b_x.value += 1
    mock.assert_called_once_with(name="b_x/A")
    ui.B.b_y += 1
    mock.assert_called_with(name="b_y/A")
    ui.B.M.m_x.value += 1
    mock2.assert_called_with(name="m_x/A")
    mock.assert_called_with(name="m_x/B")
    ui.B.M.m_y += 1
    mock2.assert_called_with(name="m_y/A")
    mock.assert_called_with(name="m_y/B")


def test_callback_block():
    @magicclass
    class A:
        f = field(int)
        result = field(str)

        @f.connect
        def _callback(self):
            self.result.value = str(self.f.value)

    ui = A()
    ui.f.value = 10
    assert str(ui.macro[-1]) == "ui.f.value = 10"
    assert ui.f.value == 10
    assert ui.result.value == "10"


def test_warning():
    # Should warn if widgets that does not have signal instance is connected with
    # callbacks
    from magicclass.widgets import Separator
    @magicclass
    class A:
        a = field(Separator)
        @a.connect
        def _callback(self):
            pass

    with pytest.warns(UserWarning):
        ui = A()

def test_container_callback():
    from magicclass.widgets import GroupBoxContainer, LineEdit
    mock = MagicMock()
    @magicclass
    class A:
        a = field(GroupBoxContainer)
        def __post_init__(self):
            self.line = LineEdit()
            self.a.append(self.line)

        @a.connect
        def _callback(self):
            mock()

    ui = A()
    mock.assert_not_called()
    ui.a[0].value = "xxx"
    mock.assert_called()

def test_parametric_callback():
    @magicclass
    class A:
        a = field(int)

        @a.connect
        def _callback(self, v):
            self._v = v

    ui = A()
    ui.a.value = 1
    assert ui._v == 1



from __future__ import annotations
from typing import Any, TypedDict
from typing_extensions import _AnnotatedAlias
from magicgui.signature import MagicSignature, split_annotated_type
from magicgui.widgets import FunctionGui
from magicgui.types import WidgetOptions
import inspect
from .utils import get_signature


class AdditionalOptions(TypedDict):
    record: bool
    keybinding: str
    into: str
    copyto: str
    moveto: str
    gui: bool


def upgrade_signature(
    func,
    gui_options: dict = None,
    caller_options: WidgetOptions = None,
    additional_options: AdditionalOptions = None,
):
    """
    Upgrade function signature to MagicMethodSignature. The input function may have
    a inspect.Signature or magicgui.signature.Magicsignature.

    Parameters
    ----------
    func : callable
        Input function.
    gui_options : dict, optional
        Options of FunctionGui.
    caller_options : WidgetOptions, optional
        Options of PushButton.
    additional_options : AdditionalOptions, optional
        Additional options that will be used in magic class.

    Returns
    -------
    callable
        Same function with upgraded signature
    """
    gui_options = gui_options or {}
    caller_options = caller_options or {}
    additional_options = additional_options or {}

    sig = get_signature(func)

    new_gui_options = MagicMethodSignature.get_gui_options(sig).copy()
    new_gui_options.update(gui_options)

    # Annotated options should also be updated
    for k, v in sig.parameters.items():
        annot = v.annotation
        if isinstance(annot, _AnnotatedAlias):
            _, widget_option = split_annotated_type(annot)
            if k in new_gui_options:
                widget_option.update(new_gui_options[k])

    new_caller_options = getattr(sig, "caller_options", {}).copy()
    new_caller_options.update(caller_options)

    new_additional_options = getattr(sig, "additional_options", {}).copy()
    new_additional_options.update(additional_options)

    func.__signature__ = MagicMethodSignature.from_signature(
        sig,
        gui_options=new_gui_options,
        caller_options=new_caller_options,
        additional_options=new_additional_options,
    )

    return func


def get_additional_option(obj: Any, option: str, default: Any = None):
    """Safely get an additional option from any objects."""
    if isinstance(obj, FunctionGui):
        sig = getattr(obj._function, "__signature__", None)
    else:
        sig = getattr(obj, "__signature__", None)
    if isinstance(sig, MagicMethodSignature):
        opt = sig.additional_options
        return opt.get(option, default)
    else:
        return default


class _void:
    """private sentinel."""


class MagicMethodSignature(MagicSignature):
    """
    This class also retains parameter options for PushButton itself, aside from the FunctionGui options
    that will be needed when the button is pushed.
    """

    def __init__(
        self,
        parameters=None,
        *,
        return_annotation=inspect.Signature.empty,
        gui_options: dict[str, dict] = None,
        caller_options: dict[str] = None,
        additional_options: dict[str] = None,
    ):
        super().__init__(
            parameters=parameters,
            return_annotation=return_annotation,
            gui_options=gui_options,
        )
        self.caller_options = caller_options or {}
        self.additional_options = additional_options or {}

    @classmethod
    def from_signature(
        cls,
        sig: inspect.Signature,
        gui_options=None,
        caller_options=None,
        additional_options=None,
    ) -> MagicMethodSignature:
        if not isinstance(sig, inspect.Signature):
            raise TypeError("'sig' must be an instance of 'inspect.Signature'")

        # prepare parameters again
        parameters = {
            k: inspect.Parameter(
                param.name,
                param.kind,
                default=param.default,
                annotation=param.annotation,
            )
            for k, param in sig.parameters.items()
        }

        return cls(
            list(parameters.values()),
            return_annotation=sig.return_annotation,
            gui_options=gui_options,
            caller_options=caller_options,
            additional_options=additional_options,
        )

    @classmethod
    def get_gui_options(cls, sig: inspect.Signature | MagicSignature) -> WidgetOptions:
        if type(sig) is inspect.Signature:
            out: WidgetOptions = {}
            for k, v in sig.parameters.items():
                annot = v.annotation
                if isinstance(annot, _AnnotatedAlias):
                    _, widget_option = split_annotated_type(annot)
                    out[k] = widget_option
            return out
        else:
            return {k: v.options for k, v in sig.parameters.items()}

    def replace(
        self,
        *,
        parameters=_void,
        return_annotation: Any = _void,
    ) -> MagicMethodSignature:
        """Create a customized copy of the Signature.

        Pass ``parameters`` and/or ``return_annotation`` arguments
        to override them in the new copy.
        """
        if parameters is _void:
            parameters = self.parameters.values()

        if return_annotation is _void:
            return_annotation = self.return_annotation
        cls = type(self)
        return cls(
            parameters,
            return_annotation=return_annotation,
            gui_options=cls.get_gui_options(self),
            caller_options=self.caller_options,
            additional_options=self.additional_options,
        )



from __future__ import annotations
from psygnal.containers import EventedList
from ._base import LayerItem


class LayerList(EventedList[LayerItem]):
    def __init__(self, data=()):
        super().__init__(data=data, hashable=False)
        self.events.removed.connect(self._on_removed)
        self.events.moved.connect(self._on_moved)

    def _on_removed(self, idx: int, value: LayerItem):
        value._visual.parent = None

    def _on_moved(self, obj: tuple[int, int]):
        src, dst = obj
        order_src = self[src]._visual.order
        order_dst = self[dst]._visual.order
        self[src]._visual.order = order_dst
        self[dst]._visual.order = order_src



class MockConfirmation:
    """Class used for confirmation test."""

    def __init__(self):
        self.text = None
        self.gui = None

    def __call__(self, text, gui):
        self.text = text
        self.gui = gui

    @property
    def value(self):
        return self.text, self.gui

    def assert_value(self, text=None, gui=None):
        if text is None and gui is None:
            assert self.text is not None and self.gui is not None
        elif text is None:
            assert self.gui is gui
        elif gui is None:
            assert self.text == text
        else:
            assert self.value == (text, gui)

    def assert_not_called(self):
        assert self.text is None and self.gui is None


def click_methods(ui):
    # TODO
    ...



from types import MethodType
from unittest.mock import MagicMock
from magicclass import magicclass, set_options, get_function_gui
from magicclass.testing import MockConfirmation
from magicgui.widgets import PushButton


def test_set_options():
    @magicclass
    class A:
        @set_options(layout="horizontal", call_button="OK", a={"widget_type": "Slider"})
        def f1(self, a: int):
            pass

        @set_options(auto_call=True)
        def f2(self, a: int):
            self.a = a

    ui = A()
    ui["f1"].changed()
    assert ui["f1"].mgui._layout == "horizontal"
    assert ui["f1"].mgui._call_button.text == "OK"
    assert ui["f1"].mgui["a"].widget_type == "Slider"

    ui["f2"].changed()
    assert ui["f2"].mgui._auto_call == True
    ui["f2"].mgui["a"].value = 2
    assert ui.a == 2
    assert str(ui.macro[-1]) == "ui.f2(a=2)"
    assert ui["f2"].mgui.visible
    ui["f2"].mgui["a"].value = 4
    assert ui.a == 4
    assert str(ui.macro[-1]) == "ui.f2(a=4)"
    assert str(ui.macro[-2]) != "ui.f2(a=2)" # for auto_call, macro should be recorded once.

def test_mgui_options():
    @magicclass
    class A:
        @set_options(layout="horizontal", auto_call=True)
        def f(self, a: int):
            self._a = a

    ui = A()
    ui["f"].changed()
    assert ui["f"].mgui.layout == "horizontal"
    ui["f"].mgui[0].value = 3
    assert ui._a == 3
    assert str(ui.macro[-1]) == "ui.f(a=3)"
    ui["f"].mgui[0].value = 5
    assert ui._a == 5
    assert str(ui.macro[-1]) == "ui.f(a=5)"
    assert str(ui.macro[-2]) != "ui.f(a=3)"

def test_annotated():
    """Test Annotated type is correctly updated."""
    from typing_extensions import Annotated
    @magicclass
    class A:
        @set_options(a={"max": 5})
        def f(self, a: Annotated[int, {"min": -5}]):
            pass

    ui = A()
    ui["f"].changed()
    assert ui["f"].mgui.a.min == -5
    assert ui["f"].mgui.a.max == 5

def test_set_design():
    from magicclass import set_design

    @magicclass
    class A:
        @set_design(text="new-text")
        def f1(self, a: int):
            pass

    ui = A()
    assert ui["f1"].text == "new-text"

def test_do_not_record():
    from magicclass import do_not_record
    @magicclass
    class A:
        @do_not_record
        def f(self): pass

    ui = A()
    ui["f"].changed()
    assert len(ui.macro) == 1

def test_nogui():
    from magicclass import nogui
    @magicclass
    class A:
        @nogui
        def f(self):
            pass

        def g(self):
            pass

    ui = A()
    assert isinstance(ui["f"], MethodType)
    assert isinstance(ui["g"], PushButton)

def test_mark_preview():
    from magicclass import mark_preview

    mock = MagicMock()
    @magicclass
    class A:
        def f0(self, x: int, y: str = "v"):
            mock(x=x, y=y)

        def f1(self, x: int = 10):
            mock(x=x)

        @mark_preview(f0, text="preview 0")
        @mark_preview(f1, text="preview 1")
        def _preview(self, x):
            mock(x=x, preview=True)

    ui = A()
    f0_gui = get_function_gui(ui, "f0")
    assert f0_gui[-2].widget_type == "PushButton"
    assert f0_gui[-2].text == "preview 0"
    f1_gui = get_function_gui(ui, "f1")
    assert f1_gui[-2].widget_type == "PushButton"
    assert f1_gui[-2].text == "preview 1"

    f0_gui[-2].changed()
    mock.assert_called_with(x=0, preview=True)
    assert str(ui.macro[-1]).startswith("#")
    mock.reset_mock()

    f0_gui[-1].changed()
    mock.assert_called_with(x=0, y="v")
    assert str(ui.macro[-1]) == "ui.f0(x=0, y='v')"
    mock.reset_mock()

    f1_gui[-2].changed()
    mock.assert_called_with(x=10, preview=True)
    assert str(ui.macro[-1]) == "ui.f0(x=0, y='v')"
    mock.reset_mock()

    f1_gui[-1].changed()
    mock.assert_called_with(x=10)
    assert str(ui.macro[-1]) == "ui.f1(x=10)"

    mock = MagicMock()

    @magicclass
    class A:
        @magicclass
        class B:
            def f0(self, x: int, y: str = "v"):
                mock(type=type(self), x=x, y=y)

            def f1(self, x: int = 10):
                mock(type=type(self), x=x)

        @mark_preview(B.f0, text="preview 0")
        @mark_preview(B.f1, text="preview 1")
        def _preview(self, x):
            mock(type=type(self), x=x, preview=True)

    ui = A()
    f0_gui = get_function_gui(ui.B, "f0")
    assert f0_gui[-2].widget_type == "PushButton"
    assert f0_gui[-2].text == "preview 0"
    f1_gui = get_function_gui(ui.B, "f1")
    assert f1_gui[-2].widget_type == "PushButton"
    assert f1_gui[-2].text == "preview 1"

    f0_gui[-2].changed()
    mock.assert_called_with(type=A, x=0, preview=True)
    assert str(ui.macro[-1]).startswith("#")
    mock.reset_mock()

    f0_gui[-1].changed()
    mock.assert_called_with(type=type(ui.B), x=0, y="v")
    assert str(ui.macro[-1]) == "ui.B.f0(x=0, y='v')"
    mock.reset_mock()

    f1_gui[-2].changed()
    mock.assert_called_with(type=A, x=10, preview=True)
    mock.reset_mock()
    assert str(ui.macro[-1]) == "ui.B.f0(x=0, y='v')"

    f1_gui[-1].changed()
    mock.assert_called_with(type=type(ui.B), x=10)
    assert str(ui.macro[-1]) == "ui.B.f1(x=10)"

def test_confirm():
    from magicclass import confirm

    # The basic usage
    mconf = MockConfirmation()

    @magicclass
    class A:
        @confirm(text="conf-text", callback=mconf)
        def f(self, a: int):
            pass

    ui = A()

    mconf.assert_not_called()
    ui.f(0)
    mconf.assert_not_called()  # no confirmation if executed programatically
    get_function_gui(ui, "f")()
    mconf.assert_value("conf-text")

    # text formating
    mconf = MockConfirmation()

    @magicclass
    class A:
        @confirm(text="<{a}>", callback=mconf)
        def f(self, a: int):
            pass

    ui = A()

    mconf.assert_not_called()
    fgui = get_function_gui(ui, "f")
    fgui()
    mconf.assert_value("<0>")
    fgui.a.value = 16
    fgui()
    mconf.assert_value("<16>")

    # test condition

    mconf = MockConfirmation()

    @magicclass
    class A:
        @confirm(text="conf-text", condition="a>5", callback=mconf)
        def f(self, a: int):
            pass

    ui = A()

    mconf.assert_not_called()
    fgui = get_function_gui(ui, "f")
    fgui()
    mconf.assert_not_called()
    fgui.a.value = 16
    fgui()
    mconf.assert_value("conf-text")

def test_confirm_with_other_wrapper():
    from magicclass import confirm

    # The basic usage
    mconf = MockConfirmation()

    @magicclass
    class A:
        @set_options(a={"max": 10})
        @confirm(text="conf-text-1", callback=mconf)
        def f1(self, a: int):
            pass

        @confirm(text="conf-text-2", callback=mconf)
        @set_options(a={"max": 12})
        def f2(self, a: int):
            pass

    ui = A()
    fgui1 = get_function_gui(ui, "f1")
    fgui2 = get_function_gui(ui, "f2")
    assert fgui1.a.max == 10
    assert fgui2.a.max == 12

    mconf.assert_not_called()
    fgui1()
    mconf.assert_value("conf-text-1")
    fgui2()
    mconf.assert_value("conf-text-2")

def test_confirm_with_thread_worker():
    from magicclass import confirm
    from magicclass.utils import thread_worker

    mconf = MockConfirmation()

    @magicclass
    class A:
        @thread_worker
        @confirm(text="conf-text-1", callback=mconf)
        def f1(self, a: int):
            pass

        @confirm(text="conf-text-2", callback=mconf)
        @thread_worker
        def f2(self, a: int):
            pass

    ui = A()
    fgui1 = get_function_gui(ui, "f1")
    fgui2 = get_function_gui(ui, "f2")
    assert isinstance(A.f1, thread_worker)
    assert isinstance(A.f2, thread_worker)

    mconf.assert_not_called()
    fgui1()
    mconf.assert_value("conf-text-1")
    fgui2()
    mconf.assert_value("conf-text-2")



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, TypeVar, Union, overload, Any
from typing_extensions import Literal

from ._gui.menu_gui import ContextMenuGui, MenuGui
from ._gui.toolbar import ToolBarGui
from ._gui.class_gui import ClassGuiBase
from ._gui._base import PopUpMode, ErrorMode, MagicTemplate

if TYPE_CHECKING:
    from .types import WidgetType, WidgetTypeStr, PopUpModeStr, ErrorModeStr
    from ._gui._function_gui import FunctionGuiPlus
    from .stylesheets import StyleSheet
    from qtpy.QtWidgets import QWidget
    from .help import HelpWidget
    from macrokit import Macro

Layout = Union[Literal["vertical"], Literal["horizontal"]]

_C = TypeVar("_C", bound=type)
_V = TypeVar("_V")

def build_help(ui: MagicTemplate, parent: QWidget | None = None) -> "HelpWidget": ...
def get_function_gui(ui: MagicTemplate, name: str) -> FunctionGuiPlus: ...
@overload
def magicclass(
    class_: _C,
    *,
    layout: Layout = "vertical",
    labels: bool = True,
    name: str | None = None,
    visible: bool | None = None,
    close_on_run: bool | None = None,
    popup_mode: PopUpModeStr | PopUpMode | None = None,
    error_mode: ErrorModeStr | ErrorMode | None = None,
    widget_type: WidgetTypeStr | WidgetType = WidgetType.none,
    icon_path: str | None = None,
    stylesheet: str | StyleSheet | None = None,
    parent=None,
) -> type[ClassGuiBase] | _C: ...
@overload
def magicclass(
    *,
    layout: Layout = "vertical",
    labels: bool = True,
    name: str | None = None,
    visible: bool | None = None,
    close_on_run: bool | None = None,
    popup_mode: PopUpModeStr | PopUpMode | None = None,
    error_mode: ErrorModeStr | ErrorMode | None = None,
    widget_type: WidgetTypeStr | WidgetType = WidgetType.none,
    icon_path: str | None = None,
    stylesheet: str | StyleSheet | None = None,
    parent=None,
) -> Callable[[_C], type[ClassGuiBase] | _C]: ...
@overload
def magicmenu(
    class_: _C,
    *,
    close_on_run: bool | None = None,
    popup_mode: str | PopUpMode | None = None,
    error_mode: str | ErrorMode | None = None,
    labels: bool = True,
    name: str | None = None,
    icon_path: str | None = None,
    parent=None,
) -> type[MenuGui] | _C: ...
@overload
def magicmenu(
    *,
    close_on_run: bool | None = None,
    popup_mode: str | PopUpMode | None = None,
    error_mode: str | ErrorMode | None = None,
    labels: bool = True,
    name: str | None = None,
    icon_path: str | None = None,
    parent=None,
) -> Callable[[_C], type[MenuGui] | _C]: ...
@overload
def magiccontext(
    class_: _C,
    *,
    close_on_run: bool | None = None,
    popup_mode: str | PopUpMode | None = None,
    error_mode: str | ErrorMode | None = None,
    labels: bool = True,
    name: str | None = None,
    icon_path: str | None = None,
    parent=None,
) -> type[ContextMenuGui] | _C: ...
@overload
def magiccontext(
    *,
    close_on_run: bool | None = None,
    popup_mode: str | PopUpMode | None = None,
    error_mode: str | ErrorMode | None = None,
    labels: bool = True,
    name: str | None = None,
    icon_path: str | None = None,
    parent=None,
) -> Callable[[_C], type[ContextMenuGui] | _C]: ...
@overload
def magictoolbar(
    class_: _C,
    *,
    close_on_run: bool | None = None,
    popup_mode: str | PopUpMode | None = None,
    error_mode: str | ErrorMode | None = None,
    labels: bool = True,
    name: str | None = None,
    icon_path: str | None = None,
    parent=None,
) -> type[ToolBarGui] | _C: ...
@overload
def magictoolbar(
    *,
    close_on_run: bool | None = None,
    popup_mode: str | PopUpMode | None = None,
    error_mode: str | ErrorMode | None = None,
    labels: bool = True,
    name: str | None = None,
    icon_path: str | None = None,
    parent=None,
) -> Callable[[_C], type[ToolBarGui] | _C]: ...
def redo(ui: MagicTemplate, index: int = -1) -> None: ...
def update_widget_state(
    ui: MagicTemplate, macro: Macro | str | None = None
) -> None: ...

class Parameters:
    def __init__(self): ...
    def __call__(self, *args: Any) -> None: ...
    def as_dict(self) -> dict[str, Any]: ...



from __future__ import annotations
import pandas as pd
from functools import wraps
from ._dataframe import WidgetDataFrame


@wraps(pd.read_csv)
def read_csv(*args, **kwargs):
    df = pd.read_csv(*args, **kwargs)
    return WidgetDataFrame(data=df)


@wraps(pd.read_excel)
def read_excel(*args, **kwargs) -> WidgetDataFrame:
    df = pd.read_excel(*args, **kwargs)
    return WidgetDataFrame(data=df)



from __future__ import annotations
import numpy as np
from vispy.visuals import Visual
from vispy.scene import transforms


class LayerItem:
    _visual: Visual

    @property
    def visible(self) -> bool:
        """Layer visibility."""
        return self._visual.visible

    @visible.setter
    def visible(self, v: bool) -> None:
        self._visual.visible = v

    def _get_transform(self) -> transforms.MatrixTransform:
        return self._visual.transform

    def affine_transform(self, mtx: np.ndarray):
        """Apply affine transformation to the layer."""
        self._visual.transform = transforms.MatrixTransform(mtx)

    @property
    def translate(self) -> np.ndarray:
        mtx = self._get_transform().matrix
        return mtx[:3, 4]



from magicclass import magicclass, magicmenu, magictoolbar, bind_key
from qtpy.QtCore import Qt
from magicclass._app import get_app
from unittest.mock import MagicMock

def test_basic_keybindings(qtbot):
    app = get_app()
    mock = MagicMock()
    from qtpy.QtTest import QTest

    @magicclass
    class A:
        @magicclass
        class B:
            @bind_key("Ctrl-T")
            def f(self): pass
            @bind_key("Ctrl-U")
            def _g(self):
                mock()

        @magicmenu
        class C:
            @bind_key("Ctrl-P")
            def f(self): pass

        @magictoolbar
        class D:
            @bind_key("Ctrl-R")
            def f(self): pass
            @bind_key("Ctrl-S")
            def _g(self):
                mock()

        @bind_key("Ctrl-A")
        def f(self): pass
        @bind_key("Ctrl-B")
        def _g(self):
            mock()

    ui = A()
    with qtbot.waitExposed(ui.native, timeout=500):
        ui.show(False)
    app.processEvents()
    QTest.keyClick(ui.native, Qt.Key_A, Qt.ControlModifier)
    app.processEvents()
    qtbot.wait(400)
    assert str(ui.macro[-1]) == "ui.f()"
    mock.assert_not_called()
    QTest.keyClick(ui.native, Qt.Key_B, Qt.ControlModifier)
    app.processEvents()
    qtbot.wait(400)
    mock.assert_called_once()
    mock.reset_mock()

    QTest.keyClick(ui.native, Qt.Key_T, Qt.ControlModifier)
    app.processEvents()
    qtbot.wait(400)
    assert str(ui.macro[-1]) == "ui.B.f()"
    mock.assert_not_called()
    QTest.keyClick(ui.native, Qt.Key_U, Qt.ControlModifier)
    app.processEvents()
    qtbot.wait(400)
    mock.assert_called_once()
    mock.reset_mock()

    QTest.keyClick(ui.native, Qt.Key_P, Qt.ControlModifier)
    app.processEvents()
    qtbot.wait(400)
    assert str(ui.macro[-1]) == "ui.C.f()"

    QTest.keyClick(ui.native, Qt.Key_R, Qt.ControlModifier)
    app.processEvents()
    qtbot.wait(400)
    assert str(ui.macro[-1]) == "ui.D.f()"
    mock.assert_not_called()
    QTest.keyClick(ui.native, Qt.Key_S, Qt.ControlModifier)
    app.processEvents()
    qtbot.wait(400)
    mock.assert_called_once()
    mock.reset_mock()



from __future__ import annotations
from functools import cached_property
import inspect
from types import MethodType
from typing import Any, TYPE_CHECKING, Callable, Iterable
import warnings
from docstring_parser import parse

if TYPE_CHECKING:
    from .._gui import BaseGui


def iter_members(cls: type, exclude_prefix: str = "__") -> Iterable[tuple[str, Any]]:
    """
    Iterate over all the members in the order of source code line number.
    This function is identical to inspect.getmembers except for the order
    of the results. We have to sort the name in the order of line number.
    """
    mro = (cls,) + inspect.getmro(cls)
    processed = set()
    names: list[str] = list(cls.__dict__.keys())
    try:
        for base in reversed(mro):
            for k in base.__dict__.keys():
                if k not in names:
                    names.append(k)

    except AttributeError:
        pass

    for key in names:
        try:
            value = getattr(cls, key)
            # handle the duplicate key
            if key in processed:
                raise AttributeError
        except AttributeError:
            for base in mro:
                if key in base.__dict__:
                    value = base.__dict__[key]
                    break
            else:
                continue
        if not key.startswith(exclude_prefix):
            yield key, value
        processed.add(key)


class Tooltips:
    """A class to manage dosctring based tooltips."""

    def __init__(self, obj: Any):
        self._doc = parse(obj.__doc__)

    @property
    def desc(self):
        doc = self._doc
        if doc.short_description is None:
            return ""
        elif doc.long_description is None:
            return doc.short_description
        else:
            return doc.short_description + "\n" + doc.long_description

    @cached_property
    def parameters(self) -> dict[str, str]:
        """Parse "Parameters" block to dict."""
        return dict(self._iter_args_of("param"))

    @cached_property
    def attributes(self) -> dict[str, str]:
        """Parse "Attributes" block to dict."""
        return dict(self._iter_args_of("attribute"))

    def _iter_args_of(self, type_name: str) -> Iterable[tuple[str, str]]:
        for p in self._doc.params:
            tp, name = p.args
            if tp == type_name:
                yield name, p.description


def get_signature(func):
    """Similar to ``inspect.signature`` but safely returns ``Signature``."""
    if hasattr(func, "__signature__"):
        sig = func.__signature__
    else:
        sig = inspect.signature(func)
    return sig


def argcount(func: Callable) -> int:
    """
    Count the number of parameters of a callable object.

    Basically, this function returns identical result as:
    >>> len(inspect.signature(func).parameters)
    but ~10x faster.
    """
    if hasattr(func, "__func__"):
        _func = func.__func__
    else:
        _func = func
    unwrapped: Callable = inspect.unwrap(
        _func, stop=(lambda f: hasattr(f, "__signature__"))
    )
    if hasattr(unwrapped, "__signature__"):
        nargs = len(unwrapped.__signature__.parameters)
    else:
        nargs = unwrapped.__code__.co_argcount
    if isinstance(func, MethodType):
        nargs -= 1
    return nargs


_LOCALS = "<locals>."


def is_instance_method(func: Callable) -> bool:
    """Check if a function is defined in a class."""
    return callable(func) and "." in func.__qualname__.split(_LOCALS)[-1]


def method_as_getter(self: BaseGui, getter: Callable):
    qualname = getter.__qualname__
    if _LOCALS in qualname:
        qualname = qualname.split(_LOCALS)[-1]
    *clsnames, funcname = qualname.split(".")
    ins = self
    self_cls = ins.__class__.__qualname__.split(".")[-1]
    if self_cls not in clsnames:
        ns = ".".join(clsnames)
        raise ValueError(
            f"Method {funcname} is in namespace {ns!r}, so it is invisible "
            f"from class {self.__class__.__qualname__!r}."
        )
    i = clsnames.index(self_cls) + 1

    for clsname in clsnames[i:]:
        ins = getattr(ins, clsname)

    def _func(w):
        return getter(ins, w)

    return _func


def eval_attribute(obj: Any, literal: str):
    attrs = literal.split(".")
    out = obj
    for attr in attrs:
        out = getattr(out, attr)
    return out


def show_tree(ui: BaseGui) -> str:
    return _get_tree(ui)


def _get_tree(ui: BaseGui, depth: int = 0):
    pref = "\t" * depth
    children_str_list: list[str] = []
    for i, child in enumerate(ui.__magicclass_children__):
        text = _get_tree(child, depth=depth + 1)
        children_str_list.append(pref + f"\t{i:>3}: {text}")

    if children_str_list:
        children_str = "\n".join(children_str_list)
        out = f"'{ui.name}'\n{children_str}"
    else:
        out = f"'{ui.name}'"
    return out


def rst_to_html(rst: str, unescape: bool = True) -> str:
    """Convert rST string into HTML."""
    from docutils.examples import html_body

    try:
        body: bytes = html_body(rst, input_encoding="utf-8", output_encoding="utf-8")
        html = body.decode(encoding="utf-8")
        if unescape:
            from xml.sax.saxutils import unescape as _unescape

            html = _unescape(html)

    except Exception as e:
        warnings.warn(
            f"Could not convert string into HTML due to {type(e).__name__}: {e}",
            UserWarning,
        )
        html = rst
    return html


def copy_info(source: type):
    def wrapper(dst):
        _copy_info(source, dst)
        return dst

    return wrapper


def _copy_info(src: type, dst: type) -> None:
    if src.__doc__:
        dst.__doc__ = src.__doc__
    for name, attr in src.__dict__.items():
        if not hasattr(dst, name) or name.startswith("_"):
            continue
        dst_attr = getattr(dst, name)
        if isinstance(attr, type):
            # common inner class
            if not isinstance(dst_attr, type):
                raise TypeError(f"{name} type mismatch.")
            _copy_info(attr, dst_attr)
        elif callable(attr):
            # common method
            if not callable(dst_attr):
                raise TypeError(f"{name} type mismatch.")
            if hasattr(attr, "__signature__"):
                dst_attr.__signature__ = attr.__signature__
            if attr.__doc__:
                dst_attr.__doc__ = attr.__doc__

    return None



"""
VTK (Visualization Toolkit) support in a magicgui/magic-class way.
This extension submodule depends on ``vedo``.

.. code-block:: shell

    pip install vedo

"""

from .widgets import VtkCanvas

__all__ = ["VtkCanvas"]



from __future__ import annotations
import numpy as np
from vispy.scene import visuals, ViewBox
from vispy.visuals import (
    VolumeVisual,
    ImageVisual,
    IsosurfaceVisual,
    MeshVisual,
)
from vispy.visuals.filters import WireframeFilter
from ._base import LayerItem

from magicgui.widgets import FloatSlider
from ...widgets import FloatRangeSlider
from ...fields import HasFields, vfield
from ...types import Color


class Image(LayerItem, HasFields):
    RENDERINGS = [
        "translucent",
        "mip",
        "minip",
        "attenuated_mip",
        "additive",
        "iso",
        "average",
    ]
    INTERPOLATIONS = ["nearest", "linear"]

    def __init__(
        self,
        data,
        viewbox: ViewBox,
        contrast_limits: tuple[float, float] | None = None,
        rendering: str = "mip",
        iso_threshold: float | None = None,
        attenuation: float = 1.0,
        cmap: str = "grays",
        gamma: str = 1.0,
        interpolation: str = "linear",
    ):
        self._data = data
        self._cache_lims()

        if contrast_limits is None:
            contrast_limits = np.min(self._data), np.max(self._data)
        if iso_threshold is None:
            iso_threshold = np.mean(contrast_limits)

        self._viewbox = viewbox
        self._create_visual(
            self._data,
            clim=contrast_limits,
            cmap=cmap,
            rendering=rendering,
            iso_threshold=iso_threshold,
            attenuation=attenuation,
            gamma=gamma,
            interpolation=interpolation,
        )

        self.contrast_limits = contrast_limits
        self.rendering = rendering
        self.iso_threshold = iso_threshold
        self.attenuation = attenuation
        self._cmap = cmap
        self.gamma = gamma
        self.interpolation = interpolation

    def _create_visual(
        self,
        data: np.ndarray,
        clim: tuple[float, float],
        cmap: str,
        rendering: str,
        iso_threshold: float,
        attenuation: float,
        gamma: float,
        interpolation: str | None = None,
    ):
        if data.ndim == 2:
            if interpolation is None:
                interpolation = "nearest"
            self._visual: ImageVisual = visuals.Image(
                data,
                clim=clim,
                method="auto",
                cmap=cmap,
                gamma=gamma,
                interpolation=interpolation,
                parent=self._viewbox.scene,
            )

        elif data.ndim == 3:
            if interpolation is None:
                interpolation = "linear"
            self._visual: VolumeVisual = visuals.Volume(
                data,
                clim=clim,
                method=rendering,
                threshold=iso_threshold,
                attenuation=attenuation,
                cmap=cmap,
                gamma=gamma,
                interpolation=interpolation,
                parent=self._viewbox.scene,
            )
        else:
            raise ValueError("Only 2D and 3D images are supported.")

    @property
    def data(self) -> np.ndarray:
        return self._data

    @data.setter
    def data(self, value) -> None:
        value = np.asarray(value)
        if self._visual is None or value.ndim != self._data.ndim:
            self._visual.parent = None
            del self._visual
            self._create_visual(value, None)
        else:
            self._visual.set_data(value)
        self._data = value
        self._cache_lims()
        self._visual.update()

    def _cache_lims(self):
        self._lims = np.min(self._data), np.max(self._data)
        self.widgets.contrast_limits.min = self._lims[0]
        self.widgets.contrast_limits.max = self._lims[1]
        self.widgets.iso_threshold.min = self._lims[0]
        self.widgets.iso_threshold.max = self._lims[1]

    # fmt: off
    rendering = vfield(str, options={"choices": RENDERINGS, "value": "mip"})
    contrast_limits = vfield(tuple[float, float], widget_type=FloatRangeSlider)
    iso_threshold = vfield(float, widget_type=FloatSlider)
    gamma = vfield(float, widget_type=FloatSlider, options={"min": 0., "max": 1.})
    attenuation = vfield(float, widget_type=FloatSlider, options={"min": 0., "max": 1.})
    interpolation = vfield(str, options={"choices": INTERPOLATIONS})
    # fmt: on

    @contrast_limits.connect
    def _on_constrast_limits_change(self, value):
        self._visual.clim = value
        self._visual.update()

    @rendering.connect
    def _on_rendering_change(self, value):
        self._visual.method = value
        self._visual.update()

    @iso_threshold.connect
    def _on_iso_threshold_change(self, value):
        self._visual.threshold = max(value, self._lims[0] + 1e-6)
        self._visual.update()

    @gamma.connect
    def _on_gamma_change(self, value):
        self._visual.gamma = value
        self._visual.update()

    @attenuation.connect
    def _on_attenuation_change(self, value):
        self._visual.attenuation = value
        self._visual.update()

    @interpolation.connect
    def _on_interpolation_change(self, value):
        self._visual.interpolation = value
        self._visual.update()


class _SurfaceBase(LayerItem):
    _visual: visuals.Mesh | visuals.Isosurface
    _wireframe: WireframeFilter

    def __init__(
        self,
        data,
        viewbox: ViewBox,
        face_color=None,
        edge_color=None,
        shading="none",
        **kwargs,
    ):
        self._viewbox = viewbox
        self._create_visual(data, **kwargs)
        self._wireframe = WireframeFilter()
        self._visual.attach(self._wireframe)
        self._data = data

        if edge_color is None:
            edge_color = [0.7, 0.7, 0.7, 1.0]
        self.edge_color = edge_color

        if face_color is None:
            face_color = [0.7, 0.7, 0.7, 1.0]
        self.face_color = face_color

        self.shading = shading

    def _create_visual(self, data):
        raise NotImplementedError()

    def _on_face_color_change(self, color):
        self._visual.color = color

    def _on_edge_color_change(self, color):
        self._wireframe.color = color
        self._visual.update()

    color = property(fget=None)

    @color.setter
    def color(self, color):
        self.face_color = color
        self.edge_color = color

    def _on_edge_width_change(self, value):
        self._wireframe.width = value
        self._visual.update()

    def _on_shading_change(self, v):
        if v == "none":
            v = None
        self._visual._shading = v


class IsoSurface(_SurfaceBase, HasFields):
    def __init__(
        self,
        data,
        viewbox: ViewBox,
        contrast_limits=None,
        iso_threshold=None,
        face_color=None,
        edge_color=None,
        shading="smooth",
    ):
        data = np.asarray(data)
        data = data.transpose(list(range(data.ndim))[::-1])

        if contrast_limits is None:
            contrast_limits = np.min(data), np.max(data)

        if iso_threshold is None:
            iso_threshold = np.mean(contrast_limits)

        super().__init__(
            data,
            viewbox=viewbox,
            face_color=face_color,
            edge_color=edge_color,
            shading=shading,
            iso_threshold=iso_threshold,
        )
        self._cache_lims()

        self.contrast_limits = contrast_limits
        self.iso_threshold = iso_threshold

    def _create_visual(self, data, iso_threshold):
        self._visual: IsosurfaceVisual = visuals.Isosurface(
            data, level=iso_threshold, parent=self._viewbox.scene
        )
        return None

    def _cache_lims(self):
        self._lims = np.min(self._data), np.max(self._data)
        self.widgets.contrast_limits.min = self._lims[0]
        self.widgets.contrast_limits.max = self._lims[1]
        self.widgets.iso_threshold.min = self._lims[0]
        self.widgets.iso_threshold.max = self._lims[1]

    @property
    def data(self) -> np.ndarray:
        return self._data

    @data.setter
    def data(self, value) -> None:
        value = np.asarray(value)
        self._visual.set_data(value)
        self._data = value
        self._visual.update()
        self._cache_lims()

    # fmt: off
    contrast_limits = vfield(tuple[float, float], widget_type=FloatRangeSlider)
    shading = vfield(str, options={"choices": ["none", "float", "smooth"], "value": "smooth"})
    iso_threshold = vfield(float, widget_type=FloatSlider)
    face_color = vfield(Color)
    edge_color = vfield(Color)
    edge_width = vfield(float, widget_type=FloatSlider, options={"min": 0.5, "max": 10.0, "value": 1.0})
    # fmt: on

    @contrast_limits.connect
    def _on_contrast_limits_change(self, value):
        self._visual.clim = value
        self._visual.update()

    @iso_threshold.connect
    def _on_iso_threshold_change(self, value):
        self._visual.level = max(self.iso_threshold, self._lims[0] + 1e-6)
        self._visual.update()

    shading.connect(_SurfaceBase._on_shading_change)
    face_color.connect(_SurfaceBase._on_face_color_change)
    edge_color.connect(_SurfaceBase._on_edge_color_change)
    edge_width.connect(_SurfaceBase._on_edge_width_change)


class Surface(_SurfaceBase, HasFields):
    def __init__(
        self,
        data,
        viewbox: ViewBox,
        face_color=None,
        edge_color=None,
        shading="none",
    ):
        super().__init__(
            data, viewbox, face_color=face_color, edge_color=edge_color, shading=shading
        )

    def _create_visual(self, data):
        self._visual: MeshVisual = visuals.Mesh(
            vertices=data[0], faces=data[1], parent=self._viewbox.scene
        )
        self.data = data

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value) -> None:
        if len(value) == 2:
            verts, faces = value
            vals = None
        elif len(value) == 3:
            verts, faces, vals = value
        else:
            raise ValueError("Data must be vertices, faces (, values).")
        self._visual.set_data(vertices=verts, faces=faces, vertex_values=vals)
        self._data = (verts, faces, vals)
        self._visual.update()

    shading = vfield(
        str, options={"choices": ["none", "float", "smooth"], "value": "smooth"}
    )
    face_color = vfield(Color)
    edge_color = vfield(Color)
    edge_width = vfield(
        float, widget_type=FloatSlider, options={"min": 0.5, "max": 10.0, "value": 1.0}
    )

    shading.connect(_SurfaceBase._on_shading_change)
    face_color.connect(_SurfaceBase._on_face_color_change)
    edge_color.connect(_SurfaceBase._on_edge_color_change)
    edge_width.connect(_SurfaceBase._on_edge_width_change)



from magicclass import magicclass, magicmenu, field, vfield, set_design
from magicclass.types import Bound
from unittest.mock import MagicMock

def test_single_wraps():
    @magicclass
    class A:
        a = vfield(int)
        @magicclass
        class B:
            def f1(self): ...
            def f2(self): ...
            def f3(self): ...
        @B.wraps
        def f2(self, a: Bound[a]):
            self.new_attr = a

    ui = A()

    # assert the widget order is conserved
    assert ui.B[0].name == "f1"
    assert ui.B[1].name == "f2"
    assert ui.B[2].name == "f3"

    mock = MagicMock()
    ui.changed.connect(mock)
    ui.B["f2"].changed()
    mock.assert_called()
    assert hasattr(ui, "new_attr")

def test_double_wrap():
    @magicclass
    class A:
        a = vfield(int)
        @magicclass
        class B:
            @magicmenu
            class C:
                def f1(self): ...
                def f2(self): ...
                def f3(self): ...
        @B.C.wraps
        def f2(self, a: Bound[a]):
            self.new_attr = a

    ui = A()

    # assert the widget order is conserved
    assert ui.B.C[0].name == "f1"
    assert ui.B.C[1].name == "f2"
    assert ui.B.C[2].name == "f3"

    ui.B.C["f2"].changed()
    assert hasattr(ui, "new_attr")

def test_copy():
    @magicclass
    class A:
        a = vfield(int)
        @magicclass
        class B:
            @magicmenu
            class C:
                def f1(self): ...
                @set_design(text="C.f2")
                def f2(self): ...
                def f3(self): ...
            @set_design(text="B.f2")
            def f2(self): ...
        @B.wraps
        @B.C.wraps
        def f2(self, a: Bound[a]):
            self.new_attr = a

    ui = A()
    assert not ui["f2"].visible
    assert ui.B["f2"].text == "B.f2"
    assert ui.B.C["f2"].text == "C.f2"

    ui.B["f2"].changed()
    assert ui.new_attr == 0
    ui.a = 1
    ui.B.C["f2"].changed()
    assert ui.new_attr == 1

def test_field_wraps():
    @magicclass
    class B:
        def f(self): ...

    @magicclass
    class A:
        b = field(B)
        a = vfield(int)

        @b.wraps
        def f(self, a: Bound[a]):
            self.new_attr = a

    ui = A()
    assert not ui["f"].visible
    ui.b["f"].changed()
    assert ui.new_attr == 0



from __future__ import annotations
from typing import Callable, Generic, Tuple, TypeVar, overload
import weakref
import vedo
import numpy as np
from .const import Representation
from ...fields import vfield, HasFields
from ...types import Color

_VtkType = TypeVar("_VtkType", bound=vedo.BaseActor)


class VtkComponent(HasFields):
    _vtk_type: type[_VtkType] | Callable[..., _VtkType]

    def __init__(
        self, *args, _parent: vedo.Plotter = None, _emit: bool = True, **kwargs
    ):
        if self._vtk_type is None:
            raise TypeError("Base VTK type is unknown.")
        self._obj = self._vtk_type(*args, **kwargs)
        self._parent_ref = weakref.ref(_parent)
        if _emit:
            self.widgets.emit_all()

    @overload
    def __init_subclass__(cls, base: type[_VtkType] = None) -> None:  # noqa
        ...

    @overload
    def __init_subclass__(cls, base: Callable[..., _VtkType] = None) -> None:  # noqa
        ...

    def __init_subclass__(cls, base=None) -> None:
        cls._vtk_type = base

    visible = vfield(True, name="visibility")

    @visible.connect
    def _on_visible_change(self, v: bool):
        if v:
            self._obj.on()
        else:
            self._obj.off()
        self._update()

    def _update(self):
        self._parent_ref().window.Render()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<{hex(id(self))}>"


class Points(VtkComponent, base=vedo.Points):
    _obj: vedo.Points
    color = vfield(Color)
    size = vfield(float)
    spherical = vfield(False)
    occlusion = vfield(float)
    pos = vfield(Tuple[float, float, float])
    scale = vfield(float)

    @color.connect
    def _on_color_change(self, v):
        self._obj.color(v[:3])
        self._update()

    @size.connect
    def _on_size_change(self, v):
        self._obj.PointSize(v)
        self._update()

    @spherical.connect
    def _on_spherical_change(self, v):
        self._obj.RenderPointsAsSpheres(v)
        self._update()

    @occlusion.connect
    def _on_occlusion_change(self, v):
        self._obj.occlusion(v)
        self._update()

    @pos.connect
    def _on_pos_change(self, v):
        self._obj.pos(*v)
        self._update()

    @scale.connect
    def _on_scale_change(self, v):
        self._obj.scale(v, absolute=True)
        self._update()


class Mesh(VtkComponent, base=vedo.Mesh):
    _obj: vedo.Mesh

    color = vfield(Color)
    occlusion = vfield(0.0)
    scale = vfield(1.0)
    representation = vfield(Representation.surface)
    linewidth = vfield(0.0, options={"min": 0.0, "max": 10})
    backface_color = vfield(Color)
    frontface_culling = vfield(False)
    backface_culling = vfield(False)
    lines_as_tubes = vfield(False)

    @color.connect
    def _on_color_change(self, v):
        self._obj.color([v * 255 for v in v[:3]])
        self._update()

    @occlusion.connect
    def _on_occlusion_change(self, v):
        self._obj.occlusion(v)
        self._update()

    @scale.connect
    def _on_scale_change(self, v):
        self._obj.scale(v, absolute=True)
        self._update()

    @representation.connect
    def _on_representation_change(self, v: Representation) -> None:
        self._obj.property.SetRepresentation(v.value)
        self._update()

    @linewidth.connect
    def _on_linewidth_change(self, v):
        self._obj.lineWidth(v)
        self._update()

    @backface_color.connect
    def _on_backface_color_change(self, v):
        self._obj.backColor([v * 255 for v in v[:3]])
        self._update()

    @frontface_culling.connect
    def _on_frontface_culling_change(self, v):
        self._obj.frontFaceCulling(v)
        self._update()

    @backface_culling.connect
    def _on_backface_culling_change(self, v):
        self._obj.backFaceCulling(v)
        self._update()

    @lines_as_tubes.connect
    def _on_lines_as_tubes_change(self, v):
        self._obj.renderLinesAsTubes(v)
        self._update()


# fmt: off
class Path(Mesh, base=vedo.Line): ...
class Sphere(Mesh, base=vedo.Sphere): ...
class Spheres(Mesh, base=vedo.Spheres): ...
class Spline(Mesh, base=vedo.Spline): ...
class KSpline(Mesh, base=vedo.KSpline): ...
class CSpline(Mesh, base=vedo.CSpline): ...
class Tube(Mesh, base=vedo.Tube): ...
class Ribbon(Mesh, base=vedo.Ribbon): ...
class Arrow(Mesh, base=vedo.Arrow): ...
class Arrows(Mesh, base=vedo.Arrows): ...
class Circle(Mesh, base=vedo.Circle): ...
class Disc(Mesh, base=vedo.Disc): ...
class Earth(Mesh, base=vedo.Earth): ...
class Ellipsoid(Mesh, base=vedo.Ellipsoid): ...
class Box(Mesh, base=vedo.Box): ...
class Cube(Mesh, base=vedo.Cube): ...
class Spring(Mesh, base=vedo.Spring): ...
class Cylinder(Mesh, base=vedo.Cylinder): ...
class Cone(Mesh, base=vedo.Cone): ...
class Text(Mesh, base=vedo.Text3D): ...
# fmt: on


def get_object_type(name: str) -> type[VtkComponent]:
    t = globals().get(name, None)
    if not isinstance(t, (type, Generic)):
        raise ValueError(f"Object type {t} not found.")
    return t



from __future__ import annotations
import sys
import logging
from contextlib import contextmanager
from qtpy import QtWidgets as QtW, QtGui, QtCore
from magicgui.backends._qtpy.widgets import QBaseWidget
from magicgui.widgets import Widget
import logging
from typing import TYPE_CHECKING, Any, Union, overload
from ..utils import rst_to_html

try:
    import numpy as np
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    from matplotlib.backends.backend_agg import FigureCanvasAgg

    FigureCanvas = FigureCanvasAgg
    MATPLOTLIB_AVAILABLE = True

except ImportError:
    MATPLOTLIB_AVAILABLE = False

if TYPE_CHECKING:
    import numpy as np
    from matplotlib.figure import Figure
    from PIL import Image
    from pathlib import Path

# See https://stackoverflow.com/questions/28655198/best-way-to-display-logs-in-pyqt


class Output:
    TEXT = 0
    HTML = 1
    IMAGE = 2


Printable = Union[str, QtGui.QImage]


class QtLogger(QtW.QTextEdit):
    process = QtCore.Signal(tuple)

    def __init__(self, parent=None, max_history: int = 500):
        super().__init__(parent=parent)
        self.setReadOnly(True)
        self.setWordWrapMode(QtGui.QTextOption.NoWrap)
        self._max_history = int(max_history)
        self._n_lines = 0
        self.process.connect(self.update)

        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)

        @self.customContextMenuRequested.connect
        def rightClickContextMenu(point):
            menu = self._make_contextmenu(point)
            if menu:
                menu.exec_(self.mapToGlobal(point))

    def update(self, output: tuple[int, Printable]):
        output_type, obj = output
        if output_type == Output.TEXT:
            self.moveCursor(QtGui.QTextCursor.End)
            self.insertPlainText(obj)
            self.moveCursor(QtGui.QTextCursor.End)
        elif output_type == Output.HTML:
            self.moveCursor(QtGui.QTextCursor.End)
            self.insertHtml(obj)
            self.moveCursor(QtGui.QTextCursor.End)
        elif output_type == Output.IMAGE:
            cursor = self.textCursor()
            cursor.insertImage(obj)
            self.insertPlainText("\n\n")
            self.verticalScrollBar().setValue(self.verticalScrollBar().maximum())
        else:
            raise TypeError("Wrong type.")
        self._post_append()

    def appendText(self, text: str):
        """Append text in the main thread."""
        self.process.emit((Output.TEXT, text))

    def appendHtml(self, html: str):
        """Append HTML in the main thread."""
        self.process.emit((Output.HTML, html))

    def appendImage(self, qimage: QtGui.QImage):
        """Append image in the main thread."""
        self.process.emit((Output.IMAGE, qimage))

    def _post_append(self):
        """Check the history length."""
        if self._n_lines < self._max_history:
            self._n_lines += 1
            return None
        cursor = self.textCursor()
        cursor.movePosition(QtGui.QTextCursor.Start)
        cursor.select(QtGui.QTextCursor.LineUnderCursor)
        cursor.removeSelectedText()
        cursor.movePosition(QtGui.QTextCursor.Down)
        cursor.deletePreviousChar()
        cursor.movePosition(QtGui.QTextCursor.End)
        self.setTextCursor(cursor)
        return None

    def _get_background_color(self) -> tuple[int, int, int, int]:
        return self.palette().color(self.backgroundRole()).getRgb()

    # These methods below are modified from qtconsole.rich_jupyter_widget.py

    def _make_contextmenu(self, pos):
        """Reimplemented to return a custom context menu for images."""
        format = self.cursorForPosition(pos).charFormat()
        name = format.stringProperty(QtGui.QTextFormat.ImageName)
        if name:
            menu = QtW.QMenu(self)

            menu.addAction("Copy Image", lambda: self._copy_image(name))
            menu.addAction("Save Image As...", lambda: self._save_image(name))
            menu.addSeparator()
            return menu

    def _copy_image(self, name):
        image = self._get_image(name)
        QtW.QApplication.clipboard().setImage(image)

    def _save_image(self, name, format="PNG"):
        """Shows a save dialog for the ImageResource with 'name'."""
        dialog = QtW.QFileDialog(self, "Save Image")
        dialog.setAcceptMode(QtW.QFileDialog.AcceptSave)
        dialog.setDefaultSuffix(format.lower())
        dialog.setNameFilter(f"{format} file (*.{format.lower()})")
        if dialog.exec_():
            filename = dialog.selectedFiles()[0]
            image = self._get_image(name)
            image.save(filename, format)

    def _get_image(self, name):
        """Returns the QImage stored as the ImageResource with 'name'."""
        document = self.document()
        image = document.resource(QtGui.QTextDocument.ImageResource, QtCore.QUrl(name))
        return image


class Logger(Widget, logging.Handler):
    """
    A widget for logging.

    Examples
    --------

    Create widget as other ``magicgui`` or ``magicclass`` widgets.

    .. code-block:: python

        logger = Logger(name="my logger")  # magicgui way

        # magicclass way
        @magicclass
        class Main:
            logger = field(Logger, name="my logger")

        # This is OK
        @magicclass
        class Main:
            logger = Logger()


    Print something in the widget

    .. code-block:: python

        # print something in the widget.
        logger.print("text")

        # a context manager that change the destination of print function.
        with logger.set_stdout():
            print("text")
            function_that_print_something()

        # permanently change the destination of print function
        sys.stdout = logger

    Logging in the widget

    .. code-block:: python

        with logger.set_logger():
            function_that_log_something()

        logging.getLogger(__name__).addHandler(logger)

    Inline plot in the widget

    .. code-block:: python

        with logger.set_plt():
            plt.plot(np.random.random(100))

    """

    current_logger: Logger | None = None

    def __init__(self):
        logging.Handler.__init__(self)
        Widget.__init__(
            self, widget_type=QBaseWidget, backend_kwargs={"qwidg": QtLogger}
        )

    def emit(self, record):
        msg = self.format(record)
        self.print(msg)

    def clear(self):
        """Clear all the histories."""
        self.native.clear()
        return None

    def print(self, *msg, sep=" ", end="\n"):
        """Print things in the end of the logger widget."""
        self.native.appendText(sep.join(map(str, msg)) + end)
        return None

    def print_html(self, html: str, end="<br></br>"):
        """Print things in the end of the logger widget using HTML string."""
        self.native.appendHtml(html + end)
        return None

    def print_rst(self, rst: str, end="\n"):
        """Print things in the end of the logger widget using rST string."""
        html = rst_to_html(rst, unescape=False)
        if end == "\n":
            end = "<br></br>"
        self.native.appendHtml(html + end)
        return None

    def print_table(
        self,
        table,
        header: bool = True,
        index: bool = True,
        precision: int | None = None,
    ):
        """
        Print object as a table in the logger widget.

        Parameters
        ----------
        table : table-like object
            Any object that can be passed to ``pandas.DataFrame`` can be used.
        header : bool, default is True
            Whether to show the header row.
        index : bool, default is True
            Whether to show the index column.
        precision: int, options
            If given, float value will be rounded by this parameter.
        """
        import pandas as pd

        df = pd.DataFrame(table)
        if precision is None:
            formatter = None
        else:
            formatter = lambda x: f"{x:.{precision}f}"
        self.native.appendHtml(
            df.to_html(header=header, index=index, float_format=formatter)
        )
        return None

    def print_image(
        self,
        arr: str | Path | np.ndarray | Image,
        vmin=None,
        vmax=None,
        cmap=None,
        norm=None,
        width=None,
        height=None,
    ) -> None:
        """Print an array as an image in the logger widget. Can be a path."""
        from magicgui import _mpl_image

        img = _mpl_image.Image()

        img.set_data(arr)
        img.set_clim(vmin, vmax)
        img.set_cmap(cmap)
        img.set_norm(norm)

        val = img.make_image()
        h, w, _ = val.shape
        image = QtGui.QImage(val, w, h, QtGui.QImage.Format_RGBA8888)

        # set scale of image
        if width is None and height is None:
            if w / 3 > h / 2:
                width = 360
            else:
                height = 240

        if width is None:
            image = image.scaledToHeight(height, QtCore.Qt.SmoothTransformation)
        else:
            image = image.scaledToWidth(width, QtCore.Qt.SmoothTransformation)

        self.native.appendImage(image)
        return None

    def print_figure(self, fig: Figure) -> None:
        """Print matplotlib Figure object like inline plot."""
        fig.canvas.draw()
        data = np.asarray(fig.canvas.renderer.buffer_rgba(), dtype=np.uint8)
        self.print_image(data)

        return None

    def write(self, msg) -> None:
        self.print(msg, end="")
        return None

    def flush(self):
        pass

    def close(self) -> None:
        # This method collides between magicgui.widgets.Widget and logging.Handler.
        # Since the close method in Widget is rarely used, here just call the latter.
        return logging.Handler.close(self)

    @contextmanager
    def set_stdout(self):
        """A context manager for printing things in this widget."""
        try:
            sys.stdout = self
            yield self
        finally:
            sys.stdout = sys.__stdout__

    @contextmanager
    def set_logger(self, name=None):
        """A context manager for logging things in this widget."""
        try:
            logging.getLogger(name).addHandler(self)
            yield self
        finally:
            logging.getLogger(name).removeHandler(self)

    @overload
    def set_plt(self, style: str | None) -> None:
        ...

    @overload
    def set_plt(self, rc_context: dict[str, Any]) -> None:
        ...

    @contextmanager
    def set_plt(self, style: str = None, rc_context: dict[str, Any] = {}):
        """A context manager for inline plot in the logger widget."""
        if not MATPLOTLIB_AVAILABLE:
            yield self
            return None
        self.__class__.current_logger = self

        if isinstance(style, dict):
            if rc_context:
                raise TypeError("style must be str.")
            rc_context = style
            style = None

        if style is None:
            style = self._get_proper_plt_style()

        backend = mpl.get_backend()
        show._called = False
        try:
            mpl.use("module://magicclass.widgets.logger")
            with plt.style.context(style), plt.rc_context(rc_context):
                yield self
        finally:
            if not show._called:
                show()
            self.__class__.current_logger = None
            mpl.use(backend)
        return None

    def _get_proper_plt_style(self) -> dict[str, Any]:
        color = self._widget._qwidget._get_background_color()[:3]
        is_dark = sum(color) < 382.5  # 255*3/2
        if is_dark:
            params = plt.style.library["dark_background"]
        else:
            keys = plt.style.library["dark_background"].keys()
            with plt.style.context("default"):
                params: dict[str, Any] = {}
                rcparams = plt.rcParams
                for key in keys:
                    params[key] = rcparams[key]
        bg = _tuple_to_color(color)
        params["figure.facecolor"] = bg
        params["axes.facecolor"] = bg
        return params


# The plt.show function will be overwriten to this.
# Modified from matplotlib_inline (BSD 3-Clause "New" or "Revised" License)
# https://github.com/ipython/matplotlib-inline
def show(close=True, block=None):
    logger = Logger.current_logger
    from matplotlib._pylab_helpers import Gcf

    try:
        for figure_manager in Gcf.get_all_fig_managers():
            logger.print_figure(figure_manager)
    finally:
        show._called = True
        if close and Gcf.get_all_fig_managers():
            plt.close("all")


def _tuple_to_color(tup: tuple[int, int, int]):
    return "#" + "".join(hex(int(t))[2:] for t in tup)



from __future__ import annotations
import sys
from typing import Union, Callable, TYPE_CHECKING

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec
from functools import wraps
from ..._gui._base import BaseGui
from ..._gui import MenuGui, ToolBarGui, ClassGui
from napari.qt.threading import GeneratorWorker, FunctionWorker


if TYPE_CHECKING:
    import napari


def to_napari(
    magic_class: type[BaseGui] | None = None,
    *,
    viewer: napari.Viewer | None = None,
):
    """
    Send magic class to current napari viewer. Classes decorated with ``@magicclass``
    ``magicmenu`` and ``magictoolbar`` are supported.
    """

    def wrapper(cls: type[BaseGui]):
        if viewer is None:
            import napari

            _viewer = napari.current_viewer()
            if _viewer is None:
                _viewer = napari.Viewer()

        if not isinstance(cls, type):
            raise TypeError(f"Cannot decorate type {type(cls)}.")

        ui = cls()
        if issubclass(cls, ClassGui):
            _viewer.window.add_dock_widget(ui)
        elif issubclass(cls, MenuGui):
            _viewer.window.main_menu.addMenu(ui.native)
            ui.native.setParent(_viewer.window.main_menu, ui.native.windowFlags())
        elif issubclass(cls, ToolBarGui):
            _viewer.window._qt_window.addToolBar(ui.native)
        else:
            raise TypeError(
                f"Class {cls.__name__} is not a magic-class. Maybe you forgot decorating"
                "the class with '@magicclass'?"
            )
        _viewer.update_console({"ui": ui})
        return cls

    if magic_class is None:
        return wrapper
    else:
        return wrapper(magic_class)


Worker = Union[FunctionWorker, GeneratorWorker]

_P = ParamSpec("_P")


def process_worker(f: Callable[_P, Worker]) -> Callable[_P, None]:
    """
    Process returned worker of ``napari.qt.threading`` in a proper way.

    Open a progress bar and start worker in a parallel thread if function is called from GUI.
    Otherwise (if function is called from script), the worker will be executed as if the
    function is directly called. This function is useful in napari because when you are
    running workers in tandem the second one does not wait for the first one to finish, which
    causes inconsistency between operations on GUI and on Python interpreter.
    """

    @wraps(f)
    def wrapper(self: BaseGui, *args, **kwargs):
        worker: Worker = f(self, *args, **kwargs)
        if self[f.__name__].running:
            worker.start()
        else:
            worker.run()
        return None

    f.__annotations__["return"] = None
    if hasattr(f, "__signature__"):
        f.__signature__ = f.__signature__.replace(return_annotation=None)
    return wrapper



from __future__ import annotations
from typing import Callable, Any
from functools import wraps
from collections import defaultdict
from qtpy.QtWidgets import QLabel, QMenu, QAction
from qtpy.QtCore import Qt
from ..utils import FreeWidget
from ...utils import Tooltips


_Callback = Callable[[Any, int], Any]


class BaseWidget(FreeWidget):
    # Abstract class for PyObject-like widgets.
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._title = QLabel(self.native)
        self._title.setAlignment(Qt.AlignCenter)
        self.running = False
        self.set_widget(self._title)
        self.title = self.name

        self._callbacks: defaultdict[type, list[_Callback]] = defaultdict(list)
        self._delegates: dict[type, Callable[[Any], str]] = dict()
        self._contextmenu: defaultdict[type, list[_Callback]] = defaultdict(list)
        self._tooltip: dict[type, Callable[[Any], str]] = dict()

    @property
    def value(self):
        raise NotImplementedError()

    @property
    def title(self) -> str:
        return self._title.text()

    @title.setter
    def title(self, text: str):
        self._title.setText(text)
        if text:
            self._title.show()
        else:
            self._title.hide()

    def register_callback(self, type_: type):
        """
        Register a double-click callback function for items of certain type.
        """

        def wrapper(func: Callable):
            self._callbacks[type_].append(func)
            return func

        return wrapper

    def register_delegate(self, type_: type):
        """
        Register a custom display.
        """

        def wrapper(func: Callable):
            self._delegates[type_] = func
            return func

        return wrapper

    def register_contextmenu(self, type_: type):
        """
        Register a custom context menu for items of certain type.
        """

        def wrapper(func: Callable):
            self._contextmenu[type_].append(func)
            return func

        return wrapper

    def register_tooltip(self, type_: type):
        """
        Register a custom tooltipfor items of certain type.
        """

        def wrapper(func: Callable | str):
            if isinstance(func, str):
                func = lambda x: func
            self._tooltip[type_] = func
            return func

        return wrapper


class ContextMenuMixin:
    """
    This class defines custom contextmenu with type dispatching.
    """

    def setContextMenu(self):
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.contextMenu)

    def setParentWidget(self, widget: BaseWidget) -> None:
        self._parent = widget
        return None

    def contextMenu(self, point):
        menu = QMenu(self)
        menu.setToolTipsVisible(True)
        item = self.itemAt(point)
        type_ = type(item.obj)
        menus = self._parent._contextmenu.get(type_, [])
        for f in menus:
            text = f.__name__.replace("_", " ")
            action = QAction(text, self)
            pfunc = partial_event(f, item.obj, self.row(item))
            action.triggered.connect(pfunc)
            doc = Tooltips(f).desc
            if doc:
                action.setToolTip(doc)
            menu.addAction(action)

        menu.exec_(self.mapToGlobal(point))


def partial_event(f, *args):
    @wraps(f)
    def _func(e):
        return f(*args)

    return _func


class PyObjectBound:
    def setObject(self, obj=None, name=None):
        if obj is not None:
            self.obj = obj
        if name is None:
            self.setText(str(obj))
        else:
            self.setText(name)
